# This file belongs to Happy Camel.
#
# Copyright (C) 2006-2008 Pieter Edelman
# (p _dot_ edelman AT gmx _dot_ net)
#
# Happy Camel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 2 only.
#
# Happy Camel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import sys, webbrowser
import main, cliparser, photohandler, config

class CLI:
  """ The command line interface. """
  
  def __init__(self, program_name, program_version, logger):
    self.logger = logger

    # Create the command line parser
    conf = config.Config()
    self.parser = self.__createParser__(conf.getValues(), program_name, program_version)
    
    # Parse the options from the command line
    self.photo_names, self.options = self.__parseCommandLine__(sys.argv[1:])
    
    # Set all options in the Config object for which it makes sense
    for option in self.options:
      try:
        conf.set(option, self.options[option])
      except config.UnknownOptionError:
        pass

    # Set the logger to the right verbosity level
    self.logger.setVerbosityLevel(self.options["verbosity"])
    
    # Create an instance of main
    self.main = main.Main(self.logger, program_name, program_version,
                          conf, self.photo_names, self.options["loc_source"],
                          self.options["utc_offset"])
      
  def getMain(self):
    """ Return the instance of Main. """
    
    return self.main
  
  def getOptions(self):
    """ Return the options dict. """
    
    return self.options
  
  def authorizeFlickr(self, url):
    """ Launch a command to authorize the Flickr-account. This is done by
        launching a webbrowser with the given URL and waiting until the user
        finishes. Return True on succes or False on failure.
        This method necessarialy breaks the verbosity level, as it has to get
        the message through. """
    
    # Spit out a message
    self.logger.message("\nYou need to authorize Happy Camel to acces your Flickr-account. Flickr achieves this in a somewhat unconventional way. When you press enter, Happy Camel launches a browser referring to a Flickr page, where you can allow or deny the permissions. Please make sure that you are logged in with the account you want to authorize.", 60)
    self.logger.message("When you are done, you need to close the browser window and press enter again.", 60)
    
    # Let the user choose a browser
    known_browsers = {}
    for browser in webbrowser._browsers.keys():
      known_browsers[len(known_browsers) + 1] = browser
    for num in range(1, len(known_browsers) + 1):
      self.logger.message("%d:\t%s" % (num, known_browsers[num]), 60)
    while (True):
      self.logger.message("\nPlease choose a browser or \"c\" to cancel (default is %s):" % webbrowser.get().name, 60)
      choice = raw_input()
      if choice.isdigit():
        if ((int(choice) > 0) and (int(choice) <= len(known_browsers))):
          browser = webbrowser.get(known_browsers[int(choice)])
          break
      elif (choice == ""):
        browser = webbrowser.get()
        break
      elif (choice == "c"):
        browser = None
        break
    
    # Fire up the webbrowser
    if (browser):
      browser.open(url)
      self.logger.message("Please press enter when you are done.", 60)
      raw_input()
      return True
    
    return False
  
  def handleFlickrAuthResult(self, result):
    """ Handle the result of the Flickr authorization process. """
    
    if (result == True):
      self.logger.message("Successfully authorized Flickr.")
    else:
      self.logger.error("There was a problem with the Flickr authorization.")
      sys.exit(1)
      
  def run(self):
    """ Run the main loop. """
    
    # Run the main program with the specified options
    options = self.getOptions()
    self.main.run(options["report_times"], options["use_geonames"],
                  options["write_exif"], options["kmz_file"],
                  options["save_config"], options["clock_diff"],
                  options["calibrate"],
                  self.authorizeFlickr, self.handleFlickrAuthResult)

  def __createParser__(self, defaults, program_name, program_version):
    """ Return the parser object given the defaults dict. """
    
    usage_str = "%prog [OPTIONS] -ACTION(S) photos"
    version_str = "%s version %s" % (program_name, program_version)
    description_str  = program_name + " is about combining GPS data with digital photos. It can do several things, like determining the photo location from a tracklog, drawing the photos in a Google Earth file"
    if (photohandler.CAN_RESIZE_PHOTOS):
      description_str += ", looking up the place names, or create a Google Earth file with the photo's at their appropriate positions along the track."
    else:
      description_str += ", or looking up the place names. To create a Google Earth file with the photos in their appropriate positions, please install the Python Imaging Library (PIL). It can be found at http://www.pythonware.com/ or installed by you're distribution if you're on Linux."
    
    parser = cliparser.CLIParser(self.logger, usage = usage_str,
                                 version = version_str,
                                 description = description_str)

    # Specifying photos
    parser.addHelpRule("Specifying photos:")
    parser.addHelpRule("Photos are specified by \"protocol://path\" or just the path for photos on disk. Photos may be located on disk (protocol \"file\") or on Flickr (protocol \"flickr\").\nFor disk photos, Unix-like wildcards are accepted.\nFlickr photos may be specified by their individual ID's (flickr://photo-id), or by set id or name using the \"sets\" prefix (flickr://sets/set-id/ or flickr://sets/set-name/). Set names only apply to sets from the logged in account and have to be unique.\n")
    
    # General options
    parser.addHelpRule("General options to control the output of this program:")
    parser.addHelpRule(parser.help_switch)
    parser.addHelpRule(parser.version_switch)
    parser.addOption("-v", "--verbosity", metavar = "verbosity",
                     var = "verbosity", type = "int",
                     default = defaults["verbosity"], help="Verbosity level. 0 is silent, 1 = errors and warnings, 2 is info, 3 is debug. Currently set to %default.")

    # Possible actions
    parser.addHelpRule("\nThe possible actions:")
    parser.addOption("-x", "--write-exif", var = "write_exif", default = False,
                     type = "bool", bool_action = True, help = "Store the coordinates in the image metadata")
    parser.addOption("-g", "--geonames", var = "use_geonames", default = False,
                     type = "bool", bool_action = True, help = "Query the Geonames web service for a location, and write that to the metadata.")
    if (photohandler.CAN_RESIZE_PHOTOS):
      parser.addOption("-k", "--kmzfile", metavar = "file", var = "kmz_file",
                       default = False, help = "Create a KMZ (Google Earth) file which displays your track and photos with the following file name.")
    parser.addOption("-r", "--report-times", var = "report_times",
                     default = False, type = "bool", bool_action = True,
                     help = "Report the begin- and end time of the track log the photo's and exit")
    parser.addOption("-s", "--save-config", var = "save_config",
                     type = "bool", default = False, bool_action = True,
                     help = "Save the options specified on this command line as default options (non-specified options are left alone).")
                      
    # The options relating to input
    parser.addHelpRule("\nThe sources of geographic information (may be more than one):") 
    parser.addOption("-t", "--tracklog", metavar = "file", var = "loc_source",
                     type = "string", action = "append", help = "Tracklog for locating the photos. Tracklogs may be compressed with GZip or BZip2, or may be stored in a Zip archive. In the latter case, all GPX files present in the archive will be read in order of presence.")
    parser.addOption("-e", "--read-exif", var = "loc_source", action = "append",
                     type = "bool", bool_action = True, help = "Use the location data already present in the image metadata.")
    
    # The options relating to Flickr
    parser.addHelpRule("\nOptions relating to communicating with Flickr:")
    parser.addOption("--flickr-user", metavar = "name", var = "flickr_user",
                     type = "string", default = defaults["flickr_user"],
                     help = "The Flickr-account which is used for reading and writing Flickr-pictures. If this is empty, don't use an account. Currently set to \"%default\".")
    parser.addOption("--flickr-sources", var = "flickr_sources",
                     type = "string", default = defaults["flickr_sources"],
                     help = "Specify which metadata sources to consider when reading images from Flickr. An \"e\" means Exif, an \"f\" means  the Flickr metadata. Multiple sources may be specified, the first one gets precedence over the second one. Currently set to \"%default\".")

    # The options relating to time correction
    parser.addHelpRule("\nOptions to align the camera and GPS time:")
    parser.addOption("--utc-offset", metavar = "timestamp", var = "utc_offset",
                     type = "string", default = None, help = "The offset from UTC (time zone and daylight savings) in which your photos were made (east is positive). Defaults to system settings at the time the recordings were made. Format is a plus or minus sign, followed by hh:mm (hours, minutes).")
    parser.addOption("-c", "--calibrate", metavar = ["file", "timestamp"],
                     nargs = 2, var = "calibrate", type = "string",
                     help = "Figure out the difference between the GPS and the camera clock using an image, and the time at which it was taken according to the GPS device (a photo of the GPS device would do nicely here). Time stamp is in the format yyyy-mm-dd-hh-mm-ss.")
    parser.addOption("--clock-difference", metavar = "seconds",
                     var = "clock_diff", type = "int", default = 0,
                     help = "The number of seconds the camera's time runs before the GPS device (note that photo's and tracklogs may or may not compensate for UTC and timezone. You may need to correct this manually. Also see -r.)")
    parser.addOption("--fix-dates", var = "fix_dates", default = defaults["fix_dates"],
                     type = "bool", bool_action = True, help = "If the clock difference is known (via --calibrate or --clock-difference), shift the times in the image metadata to their proper value. Currently set to %default")
                    
    # The options relating to interpolation
    parser.addHelpRule("\nOptions relating to interpolation of the location:")
    parser.addOption("--elevation", metavar = "mode", var = "elevation_mode",
                     type = "int", default = defaults["elevation_mode"],
                     help = "Elevation mode: 0 = none, 1 = only if known exactly, 2 = interpolate. Currently set to %default.")
    parser.addOption("--use-outside", var = "use_outside", type = "bool",
                     bool_action = True, default = defaults["use_outside"],
                     help = "If the photo's are made before the GPS recordings began or after they ended, use respectively the begin and end of the track as positions (instead of ignoring these photo's). Currently set to %default.")
    parser.addOption("--max-deviation", metavar = "meters",
                     var = "max_deviation", type = "int",
                     default = defaults["max_deviation"], help = "The maximum distance in meters an interpolated value in between tracks may be away from their edges (-1 for infinity). Currently set to %default.")

    # The options relating to Geonames                    
    parser.addHelpRule("\nOptions relating to querying the Geonames database:")                     
    parser.addOption("--geoname-tags", metavar = "tag type", var = "gn_storage",
                     type = "string", default = defaults["gn_storage"],
                     help = "The type of metadata the Geonames information needs to be stored in. k = IPTC Keywords, l = IPTC Location. This option may be combined. Currently set to \"%default\".")
    parser.addOption("--geo-info", metavar = "information type",
                      var = "gn_info", type = "string",
                      default = defaults["gn_info"], help = "The type of information the Geonames service needs to be queried for. n = Nation, c = City, s = Sub-city. These flags may be combined. Currently set to \"%default\".")
    parser.addOption("--radius", metavar = "kilometers", var = "gn_radius",
                     type = "int", default = defaults["gn_radius"],
                     help = "The search radius for the Geonames query in kilometers. Currently set to %default.")
                      
    # The options relating to KMZ file writing                    
    if (photohandler.CAN_RESIZE_PHOTOS):
      parser.addHelpRule("\nOptions relating to writing a .KMZ file:")
      parser.addOption("--sort", var = "order", default = defaults["order"],
                       help = "Sort order of the images in the KMZ file. D = Date and time, A = lAttitude, O = lOngitude, their lowercase counterparts make a reverse sort. Letters may be combined (like AOd) if you get a kick out of it. Currently set to \"%default\".")
      parser.addOption("--image-size", metavar = "pixels", var = "image_size",
                       type = "int", default = defaults["image_size"],
                       help = "The size of the long side of the photo in the KMZ file. Currently set to %default.")
      parser.addOption("--title", metavar = "title", var = "title",
                       default = defaults["title"], help = "The title of your KMZ file. Currently set to \"%default\".")
      parser.addOption("--clickable", var = "clickable", type = "bool",
                       bool_action = True, default = defaults["clickable"],
                       help = "It true, make the images clickable so they are shown in your web browser. WARNING: When using disk photos, this is not portable across computers, unless the photo's are placed at exactly the same path as on the original machine. Currently set to %default.")                      
      parser.addOption("--track", type = "bool", bool_action = True,
                       var = "show_track", default = defaults["show_track"],
                       help = "Show the GPS track in the KMZ file. Currently set to %default.")
      parser.addOption("--track-color", metavar = "color", var = "track_color",
                       default = defaults["track_color"], help = "The color of the track in 8 hexadecimal numbers (opacity, blue, green, red). Currently set to %default.")
      parser.addOption("--track-width", metavar = "pixels", var = "track_width",
                       type = "float", default = defaults["track_width"],
                       help = "The width of the track in pixels. Currently set to %1.1f." % defaults["track_width"])
      parser.addOption("--between-tracks", type = "bool", bool_action = True,
                       var = "show_between_tracks",
                       default = defaults["show_between_tracks"],
                       help = "Show lines in between the GPS tracks in the KMZ file. Currently set to %default.")         
      parser.addOption("--between-color", metavar = "color",
                       var = "between_color",
                       default = defaults["between_color"],
                       help = "Show the lines in between tracks, thus where your GPS device lost connection, with the following color. Format is the same as with --color. Currently set to %default.")
      parser.addOption("--between-width", metavar = "pixels",
                       var = "between_width", type="float",
                       default = defaults["between_width"],
                       help = "Show the lines in between tracks with the following pixel width. Currently set to %1.1f." % defaults["between_width"])                      
      parser.addOption("--waypoints", type = "bool", bool_action = True,
                       var = "show_waypoints",
                       default = defaults["show_waypoints"],
                       help = "Show the waypoints in the KMZ file. Currently set to %default.")

    # The options relating to image parsing
    parser.addHelpRule("\nUsually, metadata is read through the internal library, but written through the external Exiftool command. Using Exiftool is much slower, but using the internal library for writing carries some risk as it is not as mature.\nThe following options influence this behavior:")
    parser.addOption("--faster", var = "write_tbp", type = "bool",
                     default = defaults["write_tbp"], bool_action = True, help = "Write metadata directly instead of using Exiftool. This method is much faster, but might produce invalid pictures (a backup is automatically created when using this option). Currently set to %default.")
    parser.addOption("--exiftool", var = "read_exiftool", type = "bool",
                     default = defaults["read_exiftool"], bool_action = True, help = "Use Exiftool to both read and write the metadata metadata. This is much slower. Currently set to %default.")

    # Specify the conflicting options
    if (photohandler.CAN_RESIZE_PHOTOS):
      parser.addConflict("not (write_exif or report_times or use_geonames or fix_dates or kmz_file or save_config)",
                         "You need to specify an action, like writing the coordinates data to exif (-x), writing city and country names to the images (-g), writing a Google Earth file (-k), reporting times (-r) or fixing the dates (-f), or you may save your configuration (-s)!")
    else:                  
      parser.addConflict("not (write_exif or report_times or use_geonames or fix_dates or save_config)",
                         "You need to specify an action, like writing the coordinates data to exif (-x), writing city and country names to the images (-g), reporting times (-r) or fixing the dates (-f), or you may save your configuration (-s)!")
    parser.addConflict("(not (report_times or save_config)) and (len(loc_source) == 0)",
                       "Choose at least one source of position data (-t and/or -e)!")
    parser.addConflict("report_times and (len(__arguments__) == 0) and ((loc_source.count(True) + loc_source.count(False)) == len(loc_source))", 
                       "Choose at least one photo or one GPX file.")
    parser.addConflict("calibrate[0] and clock_diff",
                       "Choose either a time difference (--clock-difference) or a calibration image (--calibrate)!")
    parser.addConflict("fix_dates and not (calibrate or clock_diff)",
                       "A clock difference is needed to fix the photo dates, either via --clock-difference or via --calibrate.")
    parser.addConflict("(calibrate[1]) and not (re.match('\d{4}-\d{2}-\d{2}-\d{2}-\d{2}-\d{2}', calibrate[1]))",
                       "Timestamp should be in the format yyyy-mm-dd-hh-mm-ss (year, month, day, hour, minute, seconds).")
    parser.addConflict("(write_tbp and read_exiftool)", "You can't combine the --faster and --exiftool switches. The first one causes all metadata operations to be done without Exiftool, the second one causes them to be done with Exiftool.")
                       
    # Check proper formatting of arguments                     
    parser.addConflict("utc_offset and not (re.match('[+-]\d{2}\:\d{2}', utc_offset))",
                       "UTC offset should be represented by a plus or minus sign, followed by mm:ss (minute, seconds).")
    parser.addConflict("gn_info and not re.search('[scnSCN]{1,3}', gn_info)",
                       "Geonames can query for information on nation (n), city (c), and/or (sub-city).")
    parser.addConflict("gn_storage and not re.search('[kKlL]{1,2}', gn_storage)",
                       "You can store the Geonames info in IPTC Loaction tags (l), IPTC Keywords (k) or a combination of both.")
    if (photohandler.CAN_RESIZE_PHOTOS):
      parser.addConflict("order and not re.search('[daoDAO]{1,3}', order)",
                         "You can specify a sort order with either date and time (d), lattitude (a), longitude (o), or a combination of these.")
                       
    # Optionally disable some features 
    if (photohandler.CAN_RESIZE_PHOTOS):
      parser.addConflict("((loc_source.count(True) + loc_source.count(False)) == len(loc_source)) and (kmz_file) and (show_track or show_between_tracks)",
                         "No tracklog specified, suppressing track writing in KMZ file.",
                         action = "setvalues",
                         values = {"show_track": False, "show_between_tracks": False})
      parser.addConflict("(not show_track) and show_between_tracks",
                         "Showing no track, so also don't show lines in between tracks.",
                         action = "setvalues",
                         values = {"show_between_tracks": False})
    
    return parser
    
  def __parseCommandLine__(self, command_line):
    """ Parse the given command line and return a list of arguments and a dict
        of options. """
    
    # Do the actual parsing
    options, args = self.parser.parse(command_line)
      
    # Set all the options related to kmz file writing to False if we cannot
    # actually write them. This way, the rest of the script can still trust that
    # the variables are there.
    if not (photohandler.CAN_RESIZE_PHOTOS):
      options["kmz_file"]       = False
      options["order"]          = False
      options["image_size"]     = False
      options["title"]          = False
      options["clickable"]      = False
      options["show_track"]     = False
      options["track_color"]    = False
      options["track_width"]    = False
      options["bool"]           = False
      options["between_color"]  = False
      options["between_width"]  = False
      options["show_waypoints"] = False

    # Check if the user specified any photo's
    if (len(args) < 1) and not (options["report_times"] or options["save_config"]):
      self.parser.printHelp("No photos specified!", 1)
      
    return [args, options]
