# This file belongs to Happy Camel.
#
# Copyright (C) 2008 Pieter Edelman
# (p _dot_ edelman AT gmx _dot_ net)
#
# Happy Camel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 2 only.
#
# Happy Camel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

""" Handler for default and user-specified options and storage of options. """

import ConfigParser, os, types

class UnknownOptionError(Exception):
  def __init__(self, option):
    Exception.__init__(self)
    self.option = option
  def __str__(self):
    return "Unknown option \"%s\"" % self.option
    
class Config:
  """ The repository for all the values and options that aren't specific for
      this run (everything except the data sources and the actions).
      Values come from three sources, hierarchically:
      - The internal defaults.
      - The config files.
      - The user specified values for this run.
      On initalization, the first two steps are loaded. It's up to the rest of
      the program to make any changes using the set() method.
      There is one special class that follows its own rules: Flickr keys. These
      only exist in the user config file and have variable names, so they are
      treated in a special way. However, in the config file they end up as
      regular keys called "Flickr user user_name = key". """

  # Mapping of internal variable names to types, sections and names in the
  # config file and to the default values.
  map = {
    # The options regarding debugging
    "verbosity": [int, "Debug", "Verbosity", 2],
    
    # The options regarding metadata reading
    "write_tbp":     [bool, "Metadata manipulation", "Faster", False],
    "read_exiftool": [bool, "Metadata manipulation", "Exiftool", False],
    
    # The options regarding Flickr
    "flickr_sources": [str, "Flickr", "Sources", "f"],
    "flickr_user"   : [str, "Flickr", "Default user", None], 
    
    # The options regarding metadata writing
    "fix_dates": [bool, "Metadata writing", "Fix dates", False],

    # The options regarding interpolation
    "elevation_mode": [int,  "Interpolation", "Elevation", 1],
    "use_outside":    [bool, "Interpolation", "Use outside", False],
    "max_deviation":  [int,  "Interpolation", "Max deviation", 15],

    # The options regarding Geonames
    "gn_storage": [str, "Geonames", "Destination", "k"],
    "gn_info":    [str, "Geonames", "Info type", "scn"],
    "gn_radius":  [int, "Geonames", "Radius", 5],

    # The options regarding the KMZ file
    "show_track":          [bool,  "KMZ File", "Show track", False],
    "track_color":         [str,   "KMZ File", "Color", "a0ffffff"],
    "track_width":         [float, "KMZ File", "Width", 4.0],
    "show_between_tracks": [bool,  "KMZ File", "Show between tracks", False],
    "between_color":       [str,   "KMZ File", "Between color", "60ffffff"],
    "between_width":       [float, "KMZ File", "Between width", 2.0],
    "show_waypoints":      [bool,  "KMZ File", "Show waypoints", False],
    "order":               [str,   "KMZ File", "Order", "D"],
    "image_size":          [int,   "KMZ File", "Pixels", 200],
    "title":               [str,   "KMZ File", "Title", "Photo report"],
    "clickable":           [bool,  "KMZ File", "Clickable", False]
  }
    
  def __init__(self):
    """ At initialization, the values are filled with their defaults and the
        config files are read (if present). """

    # Internally, everything is stored in the values dict, where the keys are
    # the same keys as in the map.
    self.values = {}
    
    # Load the defaults from the map
    for option in self.map:
      self.values[option] = self.map[option][3]

    # Read the possible config files. FIXME: not really cross-platform.
    self.conf = ConfigParser.SafeConfigParser()
    success = self.conf.read(["/etc/happycamel.cfg", os.path.expanduser("~/.happycamel.cfg")])

    # Extract all the options from the config
    for option in self.map:
      type, section, name = self.map[option][:3]
      if (type == types.StringType):
        self.values[option] = self.__getFile__(self.values[option], section, name)
      elif (type == types.BooleanType):
        self.values[option] = self.__getFileBool__(self.values[option], section, name)
      elif (type == types.IntType):
        self.values[option] = self.__getFileInt__(self.values[option], section, name)
      elif (type == types.FloatType):
        self.values[option] = self.__getFileFloat__(self.values[option], section, name)
    
    # Make sure we have a proper config file
    if not (os.path.expanduser("~/.happycamel.cfg") in success):
      self.writeDefaults()
    else:
      self.__write__() # Just in case, maybe only for setting the access rights
      
  def get(self, option):
    """ Return the value for the option, or raise an UnknownOptionError if the
        option is unknown. """
    
    if (option in self.values):
      return self.values[option]
    else:
      raise UnknownOptionError(option)
  
  def set(self, option, value):
    """ Set the option option to value value, or raise an UnknownOptionError if
        the option is unknown. """
    
    if (option in self.map):
      if (type(value) == self.map[option][0]):
        self.values[option] = value
      else:
        raise TypeError, "Expected %s, but got %s instead" % (self.map[options][0], type(value))
    else:
      raise UnknownOptionError(option)

  def getValues(self):
    """ Return a dict with the variable names and their associated values. """
    
    return self.values
    
  def getFlickrKey(self, user_name):
    """ Returns the Flickr key for the specified user, or None if we don't have
        it. """
    
    # We have to perform our search all lowercase, to make it case insensitive
    search_option = ("flickr key %s" % user_name).lower()
    
    if (self.conf.has_section("Flickr")):
      options = self.conf.options("Flickr")
      for option in options:
        if (option.lower() == search_option):
          return self.__getFile__(None, "Flickr", option)
    return None
  
  def setFlickrKey(self, user_name, key):
    """ Set the key for the specified Flickr user. """
    
    option = "Flickr key %s" % user_name
    self.__setFile__("Flickr", option, key)

  def writeFlickrKeys(self):
    """ Write the known Flickr keys to disk. """
    
    # Since Flickr keys are already stored in the ConfigParser object, we can
    # simply Write everything to disk.
    self.__write__()
    
  def writeDefaults(self):
    """ Write the currently loaded values as the default values to disk. """
    
    # Map all the options to the ConfigParser object
    for option in self.map:
      type, section, name = self.map[option][:3]
      self.__setFile__(section, name, str(self.values[option]))
    
    # Write everything to disk
    self.__write__()

  def __getFile__(self, curr_val, section, option, method = None):
    """ Wrapper method for ConfigParser.get(). It retreives the option from the
        specified section, or returns the curr_val if it doesn't exist.
        Method is a ConfigParser get-like method. """

    # If no method is specified, use get()
    if (not method):
      method = self.conf.get

    # Try to load the value or return the current one
    try:
      return method(section, option)
    except (ConfigParser.NoSectionError, ConfigParser.NoOptionError):
      return curr_val

  def __getFileInt__(self, var, section, option):
    """ Wrapper method for ConfigParser.getint(), similar to __get__(). """
    return self.__getFile__(var, section, option, self.conf.getint)

  def __getFileBool__(self, var, section, option):
    """ Wrapper method for ConfigParser.getboolean(), similar to __get__(). """
    return self.__getFile__(var, section, option, self.conf.getboolean)

  def __getFileFloat__(self, var, section, option):
    """ Wrapper method for ConfigParser.getfloat(), similar to __get__(). """
    return self.__getFile__(var, section, option, self.conf.getfloat)

  def __setFile__(self, section, option, value):
    """ Set the option option in section section to value value in the
        ConfigParser object. """

    if not (self.conf.has_section(section)):
      self.conf.add_section(section)
    self.conf.set(section, option, value)

  def __write__(self):
    """ Write the configuration file to disk. """

    # The name of the file we want to write to
    file_name = os.path.expanduser("~/.happycamel.cfg")

    # We have to make sure that the file is only accessible by the user before
    # we starting writing secret tokens and such. So if it doesn't exist we
    # create it first. WARNING: Might not work well on Windows.
    if not os.path.exists(file_name):
      open(file_name, "w")
    os.chmod(file_name, 0600)

    # Write the configuration to the file
    self.conf.write(open(file_name, "w"))
