# This file belongs to Happy Camel.
#
# Copyright (C) 2006-2008 Pieter Edelman (p _dot_ edelman AT gmx _dot_ net)
#
# Happy Camel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 2 only.
#
# Happy Camel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

""" Library for communicating with the FLickr API. """

# Variables specifying the behaviour
# The number of seconds to wait for an answer from Flickr before calling it off:
SOCKET_TIMEOUT = 30
# The number of attempts for each call that fails because of network troubles:
NUM_ATTEMPTS = 3

import urllib, urllib2, socket, md5, re, exceptions, threading

# Sometimes Flickr isn't responding, causing the call to hang. So let's set a
# timeout.
socket.setdefaulttimeout(SOCKET_TIMEOUT)

# Module variables to define read, write and delete permissions.
class Permission:
  """ Abstract class for manipulating Flickr permissions. It allows for
      comparing different permission levels. """

  def __init__(self, repr, smaller):
    self.repr    = repr
    self.smaller = smaller
  def __repr__(self):
    return self.repr
  def __str__(self):
    return self.repr
  def __cmp__(self, other):
    if (str(other) == str(self)):
      return 0
    elif (str(other) in self.smaller):
      return 1
    return -1

READ   = Permission("read", [])
WRITE  = Permission("write", ["read"])
DELETE = Permission("delete", ["read", "write"])

# The exceptions we use
class WrongUserError(exceptions.Exception):
  """ Exception when authenticating the wrong user. """
  
  def __init__(self, found_user, expected_user):
    exceptions.Exception.__init__(self)
    self.found_user    = found_user
    self.expected_user = expected_user
    
  def __str__(self):
    return "Got authentication for user \"%s\", but excepted user \"%s\"." % (self.found_user, self.expected_user)

class NoPermissionsError(exceptions.Exception):
  """ Exception raised when there aren't enough permissions for the
      specified operation. """

  def __str__(self):
    return "Not enough permissions for this operation."

class NoTokenError(exceptions.Exception):
  """ Exception for when we haven't authenticated yet. """

  def __str__(self):
    return "I don't have permissions for this request. Please call obtainToken() first."
    
class FlickrFailure(exceptions.Exception):
  def __init__(self, num,  description):
    exceptions.Exception.__init__(self)
    self.num         = int(num)
    self.description = description
  def __str__(self):
    return "Error %d occurred when communicating with Flickr: %s" % (self.num, self.description)
    
class FlickrAPI:
  def __init__(self, api_key, secret, user_name = None, permissions = READ,
               auth_method = None, result_method = None, token = None):
    """ Initialize the Flickr API with the api_key and the secret for Flickr.
        user_name is a string describing the user name of the Flickr account
        used to perform the operations. If it is None, calls will usually not
        be authenticated.
        permissions is the level of rights we need when doing authenticated
        calls.
        auth_method is a callable function that should take an URL for
        authentication, launch a webbrowser with it, and return either True or
        False on success or failure. result_method is a callable function that
        should take handle the result of the authentication process by taking
        True or False as an argument.
        token is the Flickr token needed for manipulating Flickr pictures. if it
        is None, one will be created. """
        
    self.api_key       = api_key
    self.secret        = secret
    self.user_name     = user_name
    self.auth_method   = auth_method
    self.result_method = result_method
    self.token         = token
    self.token_perms   = None
    self.setPermissions(permissions)
    self.lock          = threading.RLock()
    
  def setPermissions(self, permissions):
    """ Set the required permission level. It needs to be a Permission class
        instance. """
    
    if (isinstance(permissions, Permission)):
      self.needed_perms = permissions
    else:
      raise TypeError, "Permission instance expected"
      
  def call(self, method, args, authenticated = None, signed = False):
    """ Call the Flickr API method with the args dict. Return a sequence of 
        True and an XML object if the method succeeds, otherwise a sequence of
        False and the error message.
        The boolean authenticated specifies whether the call should be mase as a
        Flickr user. The default behaviour is that all calls are authenticated
        when a user name is set, and not when there's no user name.
        The signed boolean specifies whether the call should be signed. Thus is
        used normally only during the authentication process. Other modules
        shouldn't need to set this value.  """
        
    # Figure out if we should make the call authenticated
    if (authenticated):
      if (not self.getUser()):
        raise "Can't make authenticated calls without an account name."
    elif (authenticated == None):
      if (self.getUser()):
        authenticated = True
      else:
        authenticated = False
    
    # Make sure we have the right permissions. Lock it in a thread-safe manner,
    # so that multithreaded applications don't all of a sudden start calling
    # this simultaniously.
    self.lock.acquire()
    if (authenticated):
      if (not self.checkTokenPermissions()):
        self.result_method(self.__obtainToken__())
    self.lock.release()
    
    # Construct the post data
    if (method[:7] != "flickr."):
      method = "flickr." + method
    args["method"] = method
    args["format"] = "json"
    post_data = self.__makePost__(args, authenticated, signed)
    
    # Make the REST request
    return self.__makeRESTRequest__("http://api.flickr.com/services/rest/", post_data)
    
  def getUser(self):
    """ Returns the user name of the logged in account. """
    
    return self.user_name
    
  def __obtainToken__(self):
    """ Obtain a token from Flickr with the permissions indicated at
        instantiation or with setPermissions(). Return True on success or False
        on failure. """
    
    # Initialize the token to None
    token = None

    # Request a frob from Flickr. This is a time-limited id that can be used to
    # construct authentication URLs
    response = self.call("flickr.auth.getFrob", {}, authenticated = False, signed = True)
    frob = response["frob"]["_content"]
    
    # Pass the authentication URL to the specified authentication method,
    # which should launch a browser with it asking the user for permissions
    # and return True when it is done.
    auth_url = "http://flickr.com/services/auth/?"
    auth_url += self.__makePost__({"perms" : self.needed_perms, "frob" : frob},
                                  authenticated = False, signed = True)
    success = self.auth_method(auth_url)
    if (success):
      # Now get the token using the just authenticated frob. This is something
      # akin to an application-specific password for the Flickr account.
      try:
        response = self.call("flickr.auth.getToken", {"frob" : frob},
                             authenticated = False, signed = True)
      except FlickrFailure, inst:
        # If the authentication through the browser failed, Flickr will respond
        # with an 108 error. In that case, we as a method have failed the
        # authentication process. When other errors occur, we simply re-raise.
        if (inst.num == 108):
          return False
        raise
      # Check if we have the correct account
      user = response["auth"]["user"]["username"]
      if (user.lower() == self.getUser().lower()):
        self.user_name = user # Set correct capitalization :S
        token = response["auth"]["token"]["_content"]
        self.token = token
        return self.checkTokenPermissions(force = True)
      else:
        raise WrongUserError(user, self.user_name)
          
    return False

  def checkTokenPermissions(self, force = False):
    """ Check the permissions of the token we have. If force is True, ignore any
        cached checks and force a response from Flikcr. """
        
    if (self.token):
      if ((self.token_perms) and not (force)):
        if (self.token_perms >= self.needed_perms): # We have it cached, so we don't need to check again
          return True
      else:
        try:
          response = self.call("flickr.auth.checkToken",
                               {"auth_token": self.token},
                               authenticated = False,
                               signed = True)
          token_perms = response["auth"]["perms"]["_content"]
          if (token_perms == "read"):
            self.token_perms = READ
          elif (token_perms == "write"):
            self.token_perms = WRITE
          elif (token_perms == "delete"):
            self.token_perms = DELETE
          if (self.token_perms >= self.needed_perms):
            return True
        except FlickrFailure, inst:
          # If we get repsonse code 98, it means the login failed or the token
          # wrong
          if (inst.num == 98):
            return False
          raise
    return False
      
  def __makeRESTRequest__(self, url, post_data, boundary = None):
    """ Make a REST request to Flickr with the url and the post data. Boundary
        is needed for multipart/form-data. """
        
    # Construct the request
    req = urllib2.Request(url)
    req.add_data(post_data)
    if (boundary):
      # We're dealing with multipart/form-data instead of 
      # application/x-www-form-urlencoded, and we should tell our request so
      req.add_header("Content-Type", "multipart/form-data; boundary=%s" % boundary)
  
    # Get a response
    attempts = 0
    while (True):
      try:
        response = urllib2.urlopen(req).read()
        break
      except:
        if (attempts < NUM_ATTEMPTS):
          attempts += 1
        else:
          raise
    
    # Parse the response
    response = re.sub("jsonFlickrApi(\(.*?\))", r"\1", response)
    response = response.replace("\/", "/")
    response = eval(response)
    
    # Return failure or success
    if (response["stat"] == "fail"):
      raise FlickrFailure(response["code"], response["message"])
    else:
      return response

  def getToken(self):
    """ Return the token. """
    
    return self.token
    
  def __makePost__(self, args, authenticated = False, signed = False):
    # Construct POST data with api key and such
     
    # Add the api_key to the args list
    args["api_key"] = self.api_key
    
    # If we need to make an authenticated call, we need to include the token
    if (authenticated):
      if (not self.token):
        raise NoTokenError
      args["auth_token"] = self.token
      
    # Calculate the signature
    if (signed) or (authenticated):
      args["api_sig"] = self.__getSignature__(args)
    
    return urllib.urlencode(args)

  def __getSignature__(self, args):
    # Calculate the signature
    keys = args.keys()
    keys.sort()
    sig_str = self.secret
    for key in keys:
      sig_str += key + str(args[key])
    return md5.new(sig_str).hexdigest()
