# This file belongs to Happy Camel.
#
# Copyright (C) 2006-2008 Pieter Edelman (p _dot_ edelman AT gmx _dot_ net)
#
# Happy Camel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 2 only.
#
# Happy Camel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

""" Module to handle GPX files and extract the relevant data from them. """

import math, time, re, bz2, gzip, zipfile, os
import xml.parsers.expat

# Regexes needed for the time parsing
seconds_re         = re.compile(".*:([0-9\.]+)[Z+-]", re.DOTALL)
utc_offset_re      = re.compile(".*(Z|[+-]\d+:\d+)", re.DOTALL)

# Regexes for detecting compressed files
compressed_openers = {"gz": gzip.GzipFile, "bz2": bz2.BZ2File, None: open}

class GPXParse:
  """ This class is essentially a handler for the parsing of GPX files. It
      cannot be called directly, but rather is a back-end for the Track class.
      """
  def __init__(self, logger, trackpoint_reporter = None,
               waypoint_reporter = None):
    """ Initialize the class with a reporter method. This needs to be a callable
        object to handle each extracted point in a tracklog. It should accept
        the parameters time, latitude, longitude, elevation, and a boolean which
        specifies if this point is the start of a (sub-)track. """
    self.logger              = logger
    self.is_gpx              = False # Is the file a valid GPX file
    self.in_track            = False # Are we in a track during processing?
    self.new_track           = False # Is this a new (sub-)track?
    self.in_waypoint         = False # Are we in a waypoint during processing?
    self.temp_content        = ""
    self.curr_time           = None
    self.curr_lat            = None
    self.curr_lon            = None
    self.curr_ele            = None
    self.curr_name           = None
    self.trackpoint_reporter = trackpoint_reporter
    self.waypoint_reporter   = waypoint_reporter

  def startElement(self, name, attributes):
    """ Called at the start of an element with element name as a string and its
        attributes as a dict. """

    # GPX element, start of a GPX file
    if (name == "gpx"):
      self.is_gpx = True

    # A track segment
    if (name == "trkseg"):
      self.in_track  = True
      self.new_track = True

    # A trackpoint
    if (self.in_track):
      if (name == "trkpt"):
        self.curr_lat = float(attributes['lat'])
        self.curr_lon = float(attributes['lon'])

    # A waypoint
    if (name == "wpt"):
      self.in_waypoint = True
      # FIXME: Doesn't this screw up track reading?
      self.curr_lat = float(attributes["lat"])
      self.curr_lon = float(attributes["lon"])
    
  def endElement(self, name):
    """ Called at the end of an element with the name. """
    
    # Track segment
    if (name == "trkseg"):
      self.in_track = False

    # Waypoint
    if (name == "wpt"):
      if (self.waypoint_reporter):
        self.waypoint_reporter(self.curr_lat, self.curr_lon, self.curr_ele, self.curr_name)
      self.in_waypoint = False
      self.curr_lat  = None
      self.curr_lon  = None
      self.curr_ele  = None
      self.curr_name = None

    if (self.in_track) or (self.in_waypoint):
      # Time
      if (name == "time"):
        self.curr_time = self.temp_content
      # Elevation
      elif (name == "ele"):
        self.curr_ele = float(self.temp_content)
      # Trackpoint
      elif (name == "trkpt"):
        if (self.trackpoint_reporter):
          self.trackpoint_reporter(self.curr_time, self.curr_lat, self.curr_lon, self.curr_ele, self.new_track)
        self.curr_time      = None
        self.curr_lat       = None
        self.curr_lon       = None
        self.curr_ele       = None
        self.new_track      = False
      # Waypoint name
      elif (name == "name"):
        self.curr_name = self.temp_content

    self.temp_content = ""

  def characters(self, content):
    """ Called when character data is encountered. """
    self.temp_content += content

class GPXFile:
  """ A gpx file with tracklog and optional waypoints on the hard disk. This 
      class provides the methods to efficiently extract the required data. """

  def __init__(self, file_path, logger, utc_offset = None, in_zip_name = None):
    """ Inititalize the GPX file with a path on the hard disk and an optional
        required offset from UTC (see setUTCOffset()). The GPX file may be GZip
        or BZ2-compressed. If in_zip_name is specified, file_path is considered
        to be the path to a ZIP file, and in_zip_name is the GPX file within the
        ZIP file. """
        
    self.logger = logger

    if (in_zip_name):
      self.compression = "zip"
      self.in_zip_name = in_zip_name
    else:
      # Figure out what kind of compression we were provided with
      self.compression = None
      for compression in ["gz", "bz2"]:
        try:
          # GZip and BZ2 raise IOError when trying to read from a non-compressed
          # file. This is used to test the compression type.
          fp = compressed_openers[compression](file_path, "r")
          fp.read(1)
          fp.close()
          self.compression = compression
          break
        except IOError:
          pass

    self.file_path = file_path
    self.setUTCOffset(utc_offset)
    
    # Vars to save the begin- and end datetimes with their locations of the 
    # tracklog. These should be None or lists of [date, lat, lon, ele].
    # All dates are stored as strings in xsd:dateTime format (as read from the
    # track). This makes them adaptable to changing UTC offsets.
    self.first_date = None
    self.last_date  = None
    
    # Check if the file is a genuine GPX file
    self.is_gpx    = False
    self.has_dates = False
    try:
      self.__parseGPX__(self.__checkForTrackPoint__)
    except Exception:
      # Everything ok
      pass
    except xml.parsers.expat.ExpatError:
      # XML Parsing error
      self.is_gpx = False
      
    if (not self.is_gpx):
      self.logger.debug("GPX file %s cannot be parsed" % self.file_path)
      raise TypeError, "File is not a proper GPX file"
    if (not self.has_dates):
      self.logger.debug("GPX file %s does not contain trackpoints with time stamps" % self.file_path)

  def setUTCOffset(self, utc_offset):
    """ Set the time zone and UTC status of the track to the required offset
        (thus this doesn't tell in which time zone the track is made). """ 
    self.utc_offset = utc_offset
  
  def getUTCOffsets(self):
    """ Return the UTC offsets (timezone + DST status) the track has to deal 
        with; the offset in which the track is recorded, and the offset that the
        user has explicitely set.
        The first response may be None if undetermined, the string "UTC" if it
        is UTC, or an integer number of seconds.
        The second response is either None or an integer  number of seconds. """
    
    # Extract the timezone from the track.
    # first_date is obtained when initializing the class. If it's None, there
    # apperently no dates in the file.
    if not (self.first_date):
      track_offset = None
    else:
      track_offset = self.parseTimeZone(self.first_date[0])
        
    return track_offset, self.utc_offset
    
  def getBeginEnd(self):
    """ Get the start- and end points of the tracklog. This method returns to
        dicts with "date" as seconds since epoch, "lat" as float in degrees,
        "lon" as float in degrees and "ele" as float in meters. """
    
    # Check if we already have these times, and if not, read them. 
    if (not (self.first_date) or not (self.last_date)):
      self.__parseGPX__(self.__searchBeginEnd__)
    
    # Make dixst out of them
    first_date = {"date": self.parseDate(self.first_date[0]),
                  "lat" : self.first_date[1],
                  "lon" : self.first_date[2],
                  "ele" : self.first_date[3]}
    last_date = {"date": self.parseDate(self.last_date[0]),
                  "lat" : self.last_date[1],
                  "lon" : self.last_date[2],
                  "ele" : self.last_date[3]}

    return (first_date, last_date)
   
  def matchTrack(self, photo_list, elevation_mode = 1, max_deviation = 0,
                 overwrite = False):
    """ Matches the times of photos in a PhotoList to the positions in a
        tracklog, interpoliting between trackpoints when needed. Elevation mode
        can be 0 (ignore elevation), 1 (use elevation if an exact poit is known)
        or 2 (interpolate). max_deviation is the maximum number of
        (interpolated) meters a photo may be from the ends of a tracklog (in 
        case of problems with the reception). overwrite specifies whether exif 
        data already present in the Photo should be overwritten. """
        
    self.logger.debug("Matching track %s." % self)
    
    # Store some vars needed by the trackparser callback method
    self.elevation_mode = elevation_mode
    self.max_deviation  = max_deviation
    self.overwrite      = overwrite
    self.photo_list     = photo_list
    
    self.curr_index = 0    # The current index in the photo list
    self.prev_point = None # __parseGPX__ needs a previous point, which is None
                           # at the start

    # Sort the photo list by date, in order to properly match it
    self.photo_list.sortByDate()

    # Make the match
    if (len(self.photo_list) > 0):
      self.__parseGPX__(self.__matchTrackPoint__)
  
  def getWaypointsAndSimpleTrack(self):
    """ Return a list of waypoints and a list of simplified tracks. The 
        waypoints are represented by a list of list containing latitude,
        longitude, elevation, and name of each waypoint. The tracks consist of 
        lists of the triplets [lat, lon, ele]. Elevation may be None in both
        cases. """

    # Initialize the vars
    self.processed_points  = 0
    self.simplified_tracks = []
    self.waypoints         = []
    self.direction  = None
    self.prev_point = None
    
    # Do the parsing
    self.__parseGPX__(self.__addSimplified__, self.__addWaypoint__)
    
    # Close the last track, if not already closed
    if ((len(self.simplified_track) > 0) and (self.prev_point)):
      if (self.simplified_track[-1][0:2] != self.prev_point[0:2]):
        self.simplified_track.append(self.prev_point)

    # Report what we've done
    track_points = 0
    for track in self.simplified_tracks:
      track_points += len(track)
    self.logger.debug("Simplified tracks with %d points to %d points." % (self.processed_points, track_points))
    return [self.waypoints, self.simplified_tracks]
  
  def __str__(self):
    """ Returns the file name of the track, without the basepath. """
    
    # Get the leaf name
    file_name = os.path.basename(self.file_path)
    
    # Properly format a track within a ZIP archive
    if (self.compression == "zip"):
      file_name = "%s (%s)" % (self.in_zip_name, file_name)

    return file_name

  def __checkForTrackPoint__(self, tp_date, lat, lon, ele, new_track):
    """ Breaks on the first proper call, abusing the standard raise mechanism. 
        Method to check quickly whether the file contains trackpoints and 
        whether these have time stamps. Also saves the first date. """
    
    self.is_gpx = True
    if ((tp_date) and (lat) and (lon)):
      self.has_dates  = True
      self.first_date = [tp_date, lat, lon, ele]
      raise Exception
    
  def __searchBeginEnd__(self, tp_date, lat, lon, ele, new_track):
    """ Search for start- and end dates of the tracks. """

    if (tp_date != None):
      if not (self.first_date):
        # It's the first date we get in our existence. Remember this moment for
        # when we're grey and old
        self.first_date = [tp_date, lat, lon, ele]
      # Live every date like it's you're last one
      self.last_date = [tp_date, lat, lon, ele]
      
  def __checkDate__(self, tp_date):
    """ Checks wether a date from a time point in a track matches the date from 
        the current photo. Returns None when no match, 1 if it matches exactly,
        2 when the photo date is between this and the previous point, -1 if the
        photo date is before the start of the tracklog. """
    
    photo_date = self.photo_list[self.curr_index].getDate()

    if (tp_date == photo_date):
      return 1
    elif (tp_date > photo_date):
      if (self.prev_point):
        return 2
      else:
        return -1
    return None
    
  def __matchTrackPoint__(self, date_str, lat, lon, ele, new_track):
    """ Callback for the GPXParse class. It matches the currently read
        trackpoint to the current photo. """

    # Match any point that has a date stamp
    if (date_str != None):
      
      # If we need to match outside the track, we should rememder the first and
      # last point
      if (not self.first_date):
        self.first_date = [date_str, lat, lon, ele]
      self.last_date = [date_str, lat, lon, ele] 
      
      # Try to match the photo at the current index
      if (self.curr_index != None): # TODO: That's the mechanism to check that we're done processing. We should find a way to actually stop it (unless we need to match outside a track).
        
        # Convert the date string
        date = self.parseDate(date_str)
        
        # Try to match the date to the first following photo in the list
        match = self.__checkDate__(date)

        # If we have a match, process it
        if (match):
          # Construct the current point as a list of latitude, longitude and
          # altitude
          curr_point = [None, None, None]
          
          if (match == 1):
            # Exact match
            self.logger.debug("Exact match for %s" % self.photo_list[self.curr_index].getName())
            curr_point = [lat, lon, ele]
          elif (match == 2):
            # Interpolate
            self.logger.debug("Interpolation for %s" % self.photo_list[self.curr_index].getName())
            fraction = (self.photo_list[self.curr_index].getDate() - self.prev_point[0]) / (date - self.prev_point[0])
            new_lat = ((lat - self.prev_point[1]) * fraction) + self.prev_point[1]
            new_lon = ((lon - self.prev_point[2]) * fraction) + self.prev_point[2]

            # Only store the point if it's an interpolation within a tracklog,
            # or if it is in the right limits when interpolating between
            # tracklogs.
            if (not (new_track)) or \
               (self.max_deviation == -1) or \
               (self.__getDiff__(self.prev_point[1], self.prev_point[2], new_lat, new_lon)[0] < self.max_deviation) or \
               (self.__getDiff__(new_lat, new_lon, lat, lon)[0] < self.max_deviation):
              curr_point[0] = new_lat
              curr_point[1] = new_lon
              
              # Check to see if we should set the elevation
              if (self.elevation_mode == 2):
                curr_point[2] = ((ele - self.prev_point[3]) * fraction) + self.prev_point[3]

            else:
              self.logger.debug("%s is in between tracks and too far away from their edges" % self.photo_list[self.curr_index].getName())
              
          elif (match == -1):
            # Photo before the start of the tracklog
            self.logger.debug("%s is made before the tracklog starts" % self.photo_list[self.curr_index].getName())
  
          # If all eleveations should be ignored, set it to None
          if (self.elevation_mode == 0):
            curr_point[2] = None

          # Pass the location to the photo object, but only if one was actually
          # found
          if ((curr_point[0] != None) and (curr_point[1] != None)):
            self.logger.debug("Found position %.6f, %.6f" % (curr_point[0], curr_point[1]))
            self.photo_list[self.curr_index].setLocation(curr_point[0], curr_point[1], curr_point[2], self.overwrite)
  
          # Raise the index by one and recursively call this method (to check the
          # next photo date with the current trackpoint)
          if (self.curr_index < (len(self.photo_list) - 1)):
            self.curr_index += 1
            self.__matchTrackPoint__(date_str, lat, lon, ele, new_track)
          else:
            self.curr_index = None
        
        # Store the current point for the next iteration  
        self.prev_point = [date, lat, lon, ele]
    
  def __addSimplified__(self, date, lat, lon, ele, new_track):
    """ This method only adds a point if it either the change in direction
        becomes larger dan 1 degree, or the difference in direction times the
        difference in position (m) becomes larger than 1. """

    if (new_track):
      # Always write the start and end of a track
      if (self.prev_point):
        if (self.prev_point[0:2] != self.simplified_track[-1][0:2]):
          self.simplified_track.append(self.prev_point)
      self.simplified_tracks.append([])
      self.simplified_track = self.simplified_tracks[-1]
      self.simplified_track.append([lat, lon, ele])
      self.direction = None
    elif (self.prev_point):
      diff, new_direction = self.__getDiff__(lat, lon, self.simplified_track[-1][0], self.simplified_track[-1][1])
      
      # Only process further if the two points actually differ
      if (diff > 0.1):
        if (self.direction):
          # Calculate the angle compared to the last stored point
          abberation = abs((new_direction - self.direction) * diff)
          # If it is different then the old one, save the old point and make
          # this the new direction
          if ((abs(new_direction - self.direction) > 0.16) or (abberation > 0.5)):
            self.simplified_track.append([lat, lon, ele])
            self.direction = new_direction
        else:
          self.direction = new_direction
          self.simplified_track.append([lat, lon, ele])
    else:
      # We're apparantly at the start of the process, so write this point
      self.simplified_track.append([lat, lon, ele])
      
    self.prev_point = [lat, lon, ele]
    self.processed_points += 1

  def __addWaypoint__(self, lat, lon, ele, name):
    """ Add a waypoint to the self.waypoints list. """
    name = name.strip()
    self.waypoints.append([lat, lon, ele, name])
    
  def __getDiff__(self, lat1, lon1, lat2, lon2):
    """ Calculate the displacements in meters between the two positions
        (roughly). """
        
    # When calculating the difference for the latitude, we need to compensate
    # for the difference in circumference of the Earth at different latitudes
    lat_diff = (math.radians(lat1 - lat2)) * 6400000 * (math.cos(math.radians(lat1)))
    
    # For the longitude, we can always use the same circumference 
    lon_diff = (math.radians(lon1 - lon2)) * 6400000
    
    diff = math.sqrt((lat_diff * lat_diff) + (lon_diff * lon_diff))
    if (diff != 0):
      direction = math.asin(lon_diff / diff)
    else:
      direction = None

    return diff, direction
    
  def __parseGPX__(self, trackpoint_reporter = None, waypoint_reporter = None):
    """ Parse the GPX file with trackpoint_reporter method as callback for each
        trackpoint read, and the waypoint_reporter method as callback for each
        waypoint read. Both methods may be None, in which case the specified
        point isn't handled. """
        
    # Create a GPXParse instance
    gpx = GPXParse(self.logger, trackpoint_reporter, waypoint_reporter)
    
    # Create an Expat parser
    parser = xml.parsers.expat.ParserCreate()
    
    # Set the callbacks
    parser.StartElementHandler  = gpx.startElement
    parser.EndElementHandler    = gpx.endElement
    parser.CharacterDataHandler = gpx.characters
    
    # Parse the file
    if (self.compression == "zip"):
      zip_fp = zipfile.ZipFile(self.file_path, "r")
      parser.Parse(zip_fp.read(self.in_zip_name))
      zip_fp.close()
    else:
      track_fp = compressed_openers[self.compression](self.file_path, "r")
      parser.ParseFile(track_fp)
      track_fp.close()

  def parseDate(self, date_str):
    """ Parse a date string in xsd:dateTime format. If the UTC offset from the 
        tracklog differs from the set offset, correct it (if no UTC offset is
        specified in the tracklog, times are left as-is. """
    
    # xsd:dateTime is always in the format
    # year-month-dayThour:minute_secondstimezone. Year, month, day, hours, and
    # minute are integers. Seconds is a float. Timezone may be omitted, have the
    # value "Z" (UTC) or may be in the form (+|-)hh:mm
    date_str = date_str.lstrip()
    year   = int(date_str[0:4])
    month  = int(date_str[5:7])
    day    = int(date_str[8:10])
    hour   = int(date_str[11:13])
    minute = int(date_str[14:16])
    seconds = re.match(seconds_re, date_str).group(1)
    # Round seconds to integer
    seconds = int(float(seconds))

    # Construct a datetime
    date = time.mktime([year, month, day, hour, minute, seconds, 0, 0, -1])
    
    timezone = self.parseTimeZone(date_str)
    if (self.utc_offset != None):
      if (timezone == "UTC"):
        # If time is in UTC we should add the given timezeone.
        date += self.utc_offset
      # Otherwise, if the time zone from the user and from the tracklog differ,
      # correct it
      elif (timezone != self.utc_offset):
        date = date - timezone + self.utc_offset
      # Otherwise, time is left alone (it's already in a time zone)
    
    return date

  def parseTimeZone(self, date_str):
    """ Parse the time zone from an xsd:dateTime format. Either return None for
        no time zone, "UTC" for UTC ("Z"), or the number of seconds. """
        
    match = re.search(utc_offset_re, date_str)
    if not (match):
      return None
    else:
      tz_str = match.group(1)

    if (tz_str == "Z"):
      return "UTC"
    else:
      tz_hours   = int(tz_str[:-3])
      tz_minutes = int(tz_str[0] + tz_str[-2:])
      tz_seconds = (tz_hours * 3600) + (tz_minutes * 60)
      return tz_seconds
    

