# This file belongs to Happy Camel.
#
# Copyright (C) 2006-2008 Pieter Edelman (p _dot_ edelman AT gmx _dot_ net)
#
# Happy Camel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 2 only.
#
# Happy Camel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

import logging

class Logger:
  def __init__(self, format = "%(message)s"):
    # This new logging mechanism of Python is quite advanced but has a bit
    # twisted interface. You can set up multiple logging channels to different
    # destinations, like the terminal, a log file, etc. Each channel (called
    # handler) has a certain verbosity level. However, they are all attached to
    # a root logger, which has it's own verbosity treshold. So messages below
    # this root treshold aren't propagated to the individual channels.
    # Therefore, we have to set it to 0.
    
    # Create a channel to the terminal
    self.output = logging.StreamHandler()
    
    # Set the output format
    self.setFormat(format)
    
    # Get the root logger (which is implicitely created) and add the terminal handler to it
    logging.getLogger().addHandler(self.output)
    
    # Set the verbosity treshold for the root logger to 0
    logging.getLogger().setLevel(0)

  def setFormat(self, format):
    """ Set the format for output messages. """
    formatter = logging.Formatter(format)
    self.output.setFormatter(formatter)
    
  def setVerbosityLevel(self, verbosity):
    """ Translate a Unix-ish verbosity level to a Python-ish level. """
    if (verbosity == 0):
      self.output.setLevel(60)
    elif (verbosity == 1):
      self.output.setLevel(30)
    elif (verbosity == 2):
      self.output.setLevel(20)
    elif (verbosity == 3):
      self.output.setLevel(10)
    
  def statement(self, msg, raw = False):
    """ Output a statement (a message about the progress of the script etc.).
        If raw is True, the message won't be formatted to the terminal width.
    """
    
    if (not raw):
      msg = self.chopNFill("[ %s ]" % msg, 80)
    logging.info(msg)
    
  def message(self, msg, level = 20, raw = False):
    """ Emit a message with a certain verbosity level. If raw is True, the
        message won't be formatted to the terminal width. """
    
    if (not raw):
      msg = self.chopNFill(msg, 80)
    logging.log(level, msg)
  
  def error(self, msg, raw = False):
    """ Emit an error. If raw is True, the message won't be formatted to the
        terminal width. """
    
    if (not raw):
      msg = self.chopNFill(msg , 80)
    logging.error(msg)

  def debug(self, msg, raw = False):
    """ Emit a debug message. If raw is True, the message won't be formatted to
        the terminal width. """
    
    if (not raw):
      msg = self.chopNFill(msg, 80)
    logging.debug(msg)
    
  def chopNFill(self, msg, width, fill = 0, first_line = ""):
    """ Format a message to the proper terminal width, breaking off preferably
        at spaces. If fill is other than zero, the message is padded left with
        the specified number of spaces. If first line is specified, this is
        inserted in the padding space at the first line. """

    # Start with first_line        
    new_msg = first_line

    # If we can fit the first line into the gutter, do so 
    if (len(first_line) <= fill):
      new_msg += (fill - len(first_line)) * " " # Pad
      line, msg = self.__breakLine__(msg, width - fill)
      new_msg += line + "\n"
    else:
      new_msg += "\n"
      
    # Iterate over all the lines of the message
    while (len(msg) > 0):
      # Pad to width
      new_msg += (fill * " ")
        
      # Strip of leading spaces
      msg = msg.lstrip()
      
      # Break the line at the proper position
      line, msg = self.__breakLine__(msg, width - fill)
      
      # Append the line
      new_msg += line + "\n"
    
    # Strip off the latest newline
    new_msg = new_msg[:-1]
    
    return new_msg
  
  def __breakLine__(self, msg, max_pos):
    """ Find a proper way to break off the text (at a space preferably). Return
        the two resulting parts. """

    if (len(msg) > max_pos):
      if ("\n" in msg[:max_pos]):
        break_start = msg.find("\n")
        break_end = break_start + 1
      else:
        break_start = msg.rfind(" ", 0, max_pos)
        if (break_start == -1):
          break_start = max_pos
        break_end = break_start
        
      return [msg[:break_start], msg[break_end:]]
    else:
      return [msg, ""]
