# This file belongs to Happy Camel.
#
# Copyright (C) 2006-2008 Pieter Edelman
# (p _dot_ edelman AT gmx _dot_ net), with contributions by Vladimir Nadvornik
# and Steffen Siebert.
#
# Happy Camel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 2 only.
#
# Happy Camel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

""" The parts for the main program flow, which can be acted upon by the user
    interface classes. """

# Import standard modules
import sys, time, os.path, types, zipfile
# Import Happy Camel modules
import gpxhandler, photohandler, kmlhandler, timehandler, flickrlib

# The Flickr API key for Happy Camel
# WARNING: This key is assigned by Flickr to Pieter Edelman ("Mr. Pi") for use
# in Happy Camel. It may not be used by anyone else for anything else than
# Happy Camel. If you would like to base your code on Happy Camel, please apply
# for your own Flickr API key (http://www.flickr.com/services/api/keys/)
FLICKR_API_KEY    = "f6440367c161bad9248168bd0bbb0010"
FLICKR_API_SECRET = "0e5b916f8545d4db"

class GeoSources:
  """ A list of geographical sources in their proper order. These can come in
      different forms, like GPXFile objects or flags indicating whether to read
      Exif data. They can also have different requirements. For example, a
      GPXFile that is used just for waypoints doesn't need to have tracklogs.
      There are methods to select lists of specific sources. """

  def __init__(self, logger):
    """ Initialize the class with the Logger instance. """

    self.logger = logger

    # Internally, all geographical sources are stored in the geo_sources list.
    # These may be GPXFile objects or a True indicating metadata matching.
    # The matchable list contains bools indicating whether the GPXFile object at
    # the correspending position can be used to match photo's, thus when it
    # contains proper tracklogs (it contains None for metadata matching).
    self.geo_sources = []
    self.matchable   = []

  def addGPXFile(self, gpx_file_path):
    """ Give a path to a GPX file (or an archive containing GPX files), load it
        and append it (or them) to the list. """

    # First, create a list of all possible paths (it may be more than one if
    # we're dealing with a ZIP archive
    gpx_paths = []

    # Test for ZIP archive
    try:
      zip_fp = zipfile.ZipFile(gpx_file_path, "r")
      self.logger.debug("Found zip file %s" % os.path.basename(gpx_file_path))
      for gpx_file in zip_fp.namelist():
        gpx_paths.append([gpx_file, gpx_file_path])
    # Otherwise, we have a single (compressed or uncompressed) track file
    except (IOError, zipfile.BadZipfile):
      gpx_paths.append([gpx_file_path, None])

    # Process all paths
    for gpx_path in gpx_paths:
      track_name = gpx_path[0]
      if (gpx_path[1]):
        track_name += " in %s" % gpx_path[1]
      self.logger.debug("Loading track %s" % track_name)
      # Try to load the track
      try:
        track = gpxhandler.GPXFile(gpx_path[0], self.logger, in_zip_name = gpx_path[1])
        self.geo_sources.append(track)
        # Set the matchable list to the proper value
        if (track.has_dates):
          self.matchable.append(True)
        else:
          self.matchable.append(False)
      except TypeError:
        self.logger.message("%s is not a GPX file - discarding" % track_name)

  def addExifInfo(self):
    """ Append a flag indicating that Exif info should be read at this
        position. """

    if not (True in self.geo_sources):
      self.geo_sources.append(True)
      self.matchable.append(None)

  def getTracks(self):
    """ Return a list of all GPXFile objects that can be used for track
        matching. """

    tracks = []

    for index in range(len(self.geo_sources)):
      if (self.matchable[index]):
        tracks.append(self.geo_sources[index])

    return tracks

  def getLocationSources(self):
    """ Return a list of all sources which can be used for location finding.
    """

    loc_sources = []

    for index in range(len(self.geo_sources)):
      geo_source = self.geo_sources[index]
      if (self.matchable[index] and isinstance(geo_source, gpxhandler.GPXFile)):
        loc_sources.append(geo_source)
      elif (geo_source == True):
        loc_sources.append(True)

    return loc_sources

  def getAllGPXFiles(self):
    """ Return a list of all GPXFile objects, whether they can be used for
        matching or not. """

    all_gpxfiles = []

    for index in range(len(self.geo_sources)):
      geo_source = self.geo_sources[index]
      if (self.matchable[index] and isinstance(geo_source, gpxhandler.GPXFile)):
        all_gpxfiles.append(geo_source)

    return all_gpxfiles
    
class Main:
  """ The main Happy Camel class. """

  def __init__(self, logger, program_name, program_version, config, photo_refs,
               loc_source_list, utc_offset):
    """ Initialize the main class with the parameters.
        logger is a Logger instance.
        program_name and program_version are strings specifying the program
        name and version.
        config is a Config instance.
        photo_refs is a list of references to photos, either paths files or
        Flickr-style URL pointing to individual photos or sets.
        loc_source_list is a list specifying paths to gpx files and booleans.
        utc_offset is an integer specifying the number of seconds the time zone
        is differs with UTC.
    """

    self.conf = config

    # Print a welcome message
    self.logger = logger
    self.logger.message("Welcome to %s version %s" % (program_name, program_version))

    # Load the tracks, and build a hierarchical list of location sources
    if ((loc_source_list.count(True) + loc_source_list.count(False)) < len(loc_source_list)):
      self.logger.statement("Loading track(s)")
    self.geo_sources = self.getLocSources(loc_source_list)

    # Save the photo names for later
    self.photo_refs = photo_refs
    
    # Set the UTC Offset
    self.setUTCOffset(utc_offset)

  def run(self, report_times, use_geonames, write_exif, kmz_file,
          save_config, clock_diff, calibrate,
          flickr_auth_method, flickr_result_method):

    """ Run the main loop, thus matching photo's, writing metadata, creating
        KMZ files, etc. """

    # Save the configuration
    if (save_config):
      self.logger.statement("Saving configuration")
      self.conf.writeDefaults()
      if not (report_times or write_exif or kmz_file):
        sys.exit(0)

    # Figure out what kind of permissions we need
    if ((use_geonames) or (write_exif) or (self.conf.get("fix_dates"))):
      permissions = flickrlib.WRITE
    else:
      permissions = flickrlib.READ

    # Initialize our FlickrAPI object
    flickr_user = self.conf.get("flickr_user")
    if (flickr_user):
      flickr_token = self.conf.getFlickrKey(flickr_user)
    else:
      flickr_token = None
    self.flickr_api = flickrlib.FlickrAPI(FLICKR_API_KEY, FLICKR_API_SECRET,
                                          flickr_user, permissions,
                                          flickr_auth_method,
                                          flickr_result_method,
                                          flickr_token)

    # Get the time delta
    time_delta = self.getTimeDelta(clock_diff, calibrate,
                                   not self.conf.get("read_exiftool"),
                                   self.conf.get("flickr_sources"))
               
    # Load the photos
    photo_list = self.getPhotoList(self.photo_refs, time_delta,
                                   kmz_file, self.geo_sources,
                                   not self.conf.get("read_exiftool"),
                                   self.conf.get("write_tbp"),
                                   use_geonames, write_exif,
                                   self.conf.get("flickr_sources"))
    
    # Check if we have enough geographical sources
    if ((len(self.geo_sources.getLocationSources()) == 0) and not (report_times)):
      self.logger.error("No valid input sources for geographical data specified!")
      sys.exit(1)

    # Report the times
    if (report_times):
      self.reportTimes(photo_list)
      sys.exit(0)

    # Match the tracklogs to the photos
    unmatched_photo_list = self.matchPhotos(photo_list, self.geo_sources,
                                            self.conf.get("elevation_mode"),
                                            self.conf.get("max_deviation"),
                                            self.conf.get("use_outside"))
    if (len(photo_list) == 0):
      self.logger.error("I couldn't find a location for any of your photo's!")

    # Query geonames
    if (use_geonames and self.conf.get("gn_info") and (len(photo_list) > 0)):
      self.logger.statement("Querying the Geonames database (geonames.org)")
      photo_list.queryGeoNames(self.conf.get("gn_radius"),
                               self.conf.get("gn_info").lower())

    # Write the metadata
    if (write_exif or use_geonames or self.conf.get("fix_dates")):
      self.logger.statement("Writing the metadata")
      self.writeMetadata(photo_list, unmatched_photo_list, write_exif,
                         self.conf.get("gn_storage"), self.conf.get("fix_dates"))

    # Write the .KMZ file
    if (photohandler.CAN_RESIZE_PHOTOS):
      if (kmz_file and (len(photo_list) > 0)):
        self.logger.statement("Writing KMZ file")
        self.writeKMZFile(self.geo_sources.getAllGPXFiles(), photo_list,
                          kmz_file, self.conf.get("order"),
                          self.conf.get("image_size"),
                          self.conf.get("title"),
                          self.conf.get("track_color"),
                          self.conf.get("track_width"),
                          self.conf.get("between_color"),
                          self.conf.get("between_width"),
                          self.conf.get("clickable"),
                          self.conf.get("show_waypoints"),
                          self.conf.get("show_track"),
                          self.conf.get("show_between_tracks"))

  def getLocSources(self, input_list):
    """ Build and return a GeoSources instance with all possible sources of
        geographical information needed, based on the input list; a list of
        paths to GPX files or archives containging GPX files, or True or False
        values indicating whether to read metadata from the images. """

    # Create a GeoSources instance
    geo_sources = GeoSources(self.logger)

    # Add everything
    for source in input_list:
      if (type(source) == types.StringType):
        geo_sources.addGPXFile(source)
      elif (source == True):
        geo_sources.addExifInfo()

    return geo_sources

  def getTimeDelta(self, user_diff = None, user_calibrate = [None, None], 
                   read_tbp = True, flickr_sources = "f"):
    """ Get the number of seconds the camera clock runs before real time, based
        on the user specified time_diff in seconds or the user specified
        caibration image and time stamp. This should be presented in the
        user_calibrate list, with a path and a time stamp in the form of
        YYYY-MM-DD-HH-MM-SS. """

    # Check if the user specified a time difference on the command line
    if (user_diff):
      time_delta = user_diff

    # Check if we need to figure out the time difference with a calibration
    # image
    elif (user_calibrate[0] and user_calibrate[1]):
      try:
        time_delta = timehandler.getPhotoTimeOffset(user_calibrate[0],
                                                    user_calibrate[1],
                                                    self.logger, read_tbp,
                                                    self.flickr_api,
                                                    flickr_sources)
      except timehandler.NoImageError:
        self.logger.error("No valid calibration image specified!")
        sys.exit(1)
    else:
      time_delta = 0

    self.logger.debug("Using a time delta of %d" % time_delta)

    return time_delta

  def getPhotoList(self, photo_refs, time_delta, kmz_file,
                   geo_sources, read_tbp, write_tbp, use_geonames, write_exif,
                   flickr_sources):
    """ Load the photos from the photo_refs list, and return a PhotoList object.
        time_delta specifies the number of seconds the camera clock runs before
        real time, kmz_file is the optional name of a KMZ file to write and
        geo_sources is the GeoSources object containing all the sources of
        geographical information. read_tbp and write_tbp specify whether to use 
        the TBP library for reading and writing of photo metadata.
        flickr_sources is a list of metadata sources
        to consider when querying Flickr, where "e" is exif and "f" is Flickr.
    """

    # Load the photo's, if the user specified them (this might not be the case
    # if she only wants to know info about the tracklog).
    photo_list = photohandler.PhotoList(photo_refs, time_delta, self.logger,
                                        read_tbp, write_tbp, self.flickr_api) # FIXME: Is time_delta still of any use here?

    if (len(photo_refs) > 0):
      # Bail out if we don't have any photo's left
      self.logger.statement("Loading photos")
      if (len(photo_list) == 0):
        self.logger.error("I couldn't find any legitimate photos!")
        sys.exit(1)

      if (len(photo_list.getByType()["disk"]) > 0) and \
        (self.conf.get("read_exiftool")) or \
        ((not self.conf.get("write_tbp")) and \
         (use_geonames or write_exif or self.conf.get("fix_dates"))):
        if not (photohandler.haveExiftool()):
          self.logger.error("You specified to use Exiftool for metadata manipulation, but you don't have it installed. Please use the correct switches (--faster and --exiftool=n) or obtain a copy of Exiftool (http://www.sno.phy.queensu.ca/~phil/exiftool/).")
          sys.exit(1)

      # Figure out whether we need to read comments, positions and additional
      # Flickr-info
      if (kmz_file):
        read_comment     = True
        read_flickr_data = True
      else:
        read_comment     = False
        read_flickr_data = False

      if (True in geo_sources.getLocationSources()): read_pos = True
      else: read_pos = False

      # Check our Flickr permissions
      for photo in photo_list:
        if (isinstance(photo, photohandler.FlickrPhoto)):
          permissions_ok = self.flickr_api.checkTokenPermissions()
          if (permissions_ok):
            self.conf.setFlickrKey(self.flickr_api.getUser(),
                                   self.flickr_api.getToken())
            self.conf.writeFlickrKeys()
          break

      # Read the necessary metadata for each photo
      success = photo_list.loadMetadata(read_date = True,
                                        read_comment = read_comment,
                                        read_pos = read_pos,
                                        read_flickr_data = read_flickr_data,
                                        time_delta = time_delta,
                                        flickr_sources = flickr_sources)
      for photo in photo_list:
        if not (success[photo]):
          self.logger.error("Unable to read metadata from %s." % photo.getName())

      # Clean up the photo list. If we need to match it to a track, the
      # criterium should be datetimes. If need to read Exif times, the criterium
      # should be their existence. These are mutually exclusive.
      if (len(geo_sources.getTracks())):
        photo_list.cleanup(check_date = True)
        if (len(photo_list) == 0):
          self.logger.error("I couldn't find a date and time for any of your photos!")
          sys.exit(1)
      elif (geo_sources.getLocationSources().count(True)):
        photo_list.cleanup(check_location = True)
        if (len(photo_list) == 0):
          self.logger.error("I couldn't find a location for any of your photos!")
          sys.exit(1)

    return photo_list

  def setUTCOffset(self, user_offset):
    """ Set the UTC offset of the GPS data, based on the specified user offset
        and the geographic sources. user_offset should be None or a string in
        the form of either a plus or minus sign, followed by two numbers
        defining the hours, followed by two numbers defining the seconds. """

    # We just need the tracks, no True or False statements
    tracks = self.geo_sources.getTracks()

    if (len(tracks)):
      # If the user specified an offset, parse it and set all the tracks to it
      if (user_offset != None):
        # Convert the string of (+/-)HH:MM to seconds as int
        self.user_offset = (int(user_offset[0:3]) * 3600) + (int(user_offset[0] + user_offset[4:6]) * 60)
        self.logger.debug("User specified offset from UTC is %d seconds." % self.user_offset)
        for track in tracks:
          track.setUTCOffset(self.user_offset)
      else:
        # Otherwise, set it to the system time zone
        self.user_offset = None

        for track in tracks:
          # Get the system UTC offset at the time the track was made
          start_time = time.localtime(track.getBeginEnd()[0]["date"])
          system_offset = timehandler.getSystemUTCOffset(self.logger, start_time)
          # Set the offset
          track.setUTCOffset(system_offset)
          self.logger.debug("Using the system's UTC offset, which is %d seconds, for track %s." % (system_offset, track))

  def reportTimes(self, photo_list):
    """ Report the times of the tracklogs and photo's to the internal Logger
        object. """

    # For each track
    for track in self.geo_sources.getTracks():
      # Start a new line
      tracklog_msg = "\n- "

      # Collect the start- and endpoints of the current track
      track_times = []
      for point in track.getBeginEnd():
        track_times.append(point["date"])

      # Format info about the time zone

      # If the track has no time zone set, we're done soon
      if (track.getUTCOffsets()[0] == None):
        tracklog_msg += "The tracklog %s does not specify a time zone and is considered to be in local time:" % track
      else:
        # Otherwise, it can be either in UTC or in a time zone
        if (track.getUTCOffsets()[0] == "UTC"):
          tracklog_msg +=  "The tracklog %s is recorded in UTC" % track
        else:
          tracklog_msg += "The tracklog %s is recorded in UTC%s%02d:%02d" % (track, timehandler.formatUTCOffset(track.getUTCOffsets()[0]))

        # If the track has an UTC offset explicitely set, it can be because of
        # a user specified offset, or because it was in UTC and is converted to
        # local time.
        if (track.getUTCOffsets()[1] != None):
          if (self.user_offset != None):
            tracklog_msg += ", but is converted to %s.\nThe corrected times become"  % timehandler.formatUTCOffset(track.getUTCOffsets()[1])
          elif (track.getUTCOffsets()[0] == "UTC"):
            tracklog_msg += ", and is converted to local time: %s (you can override this with the --utc-offset option).\nThe corrected times become"  % timehandler.formatUTCOffset(track.getUTCOffsets()[1])

      # Report the corrected times
      tracklog_msg += ":\nStart: %s,  End: %s" % (time.strftime("%d/%m/%Y, %H:%M:%S", time.localtime(track_times[0])), time.strftime("%d/%m/%Y, %H:%M:%S", time.localtime(track_times[-1])))
      self.logger.message(tracklog_msg, level = 50)

    # Report the photo times
    if (len(photo_list) > 0):
      photo_times = photo_list.getBeginEnd()

      # Format the time string
      # Calculate days, hours, minutes, seconds
      seconds = abs(photo_list.getTimeDelta()) % 60
      minutes = abs(photo_list.getTimeDelta()) / 60
      hours = minutes / 60
      minutes = minutes % 60
      days = hours / 24
      hours = hours % 24

      # Buid the time string
      time_str = "%d second%s" % (seconds, (seconds <> 1) and "s" or "")
      if (minutes or hours or days):
        time_str = "%d minute%s and %s" % (minutes, ((minutes <> 1) and "s" or ""), time_str)
        if (hours or days):
          time_str = "%d hour%s, %s" % (hours, ((hours <> 1) and "s" or ""), time_str)
          if (days):
            time_str = "%d day%s, %s" % (days, ((days <> 1) and "s" or ""), time_str)

      # Add the time direction
      if (photo_list.getTimeDelta() < 0):
        time_str += " behind"
      else:
        time_str += " ahead of"
      self.logger.message("\n- The camera clock runs %s GPS time. (You can change this with the --calibrate switch or with the --clock-difference option.)" % time_str)
      self.logger.message("The corrected times become:\nStart: %s,  End: %s" % (time.strftime("%d/%m/%Y, %H:%M:%S", time.localtime(photo_times[0])), time.strftime("%d/%m/%Y, %H:%M:%S", time.localtime(photo_times[-1]))), level = 50)

  def matchPhotos(self, photo_list, loc_sources, elevation_mode, max_deviation,
                  use_outside):
    """ Match the photo's in photo_list to the location sources in loc_sources,
        clean up the photo list and return a list of unmatched photo's. """

    # Do the matching
    photo_list.matchTracks(loc_sources.getLocationSources(), elevation_mode,
                           max_deviation, use_outside)

    # Clean up the photo list
    unlocated_photo_list = photo_list.cleanup(check_location = True)

    return unlocated_photo_list

  def writeMetadata(self, photo_list, unmatched_photo_list, write_exif,
                    gn_storage, fix_dates):
    """ Write the obtained metadata and fixed dates to the photo's in photo_list
        and the fixed dates to unmatched_photo_list. """

    # Write the location, geonames info and fixed datetimes
    matched_results = photo_list.writeMetadata(write_exif, gn_storage, fix_dates)
    
    # Write the fixed datetimes for unmatched photo's
    unmatched_results = unmatched_photo_list.writeMetadata(False, False, True)
    
    # Report the results
    results = matched_results
    results.update(unmatched_results)
    for photo in results:
      if not (results[photo]):
        self.logger.message("Couldn't succesfully write metadata to photo %s." % photo.getName())

  def writeKMZFile(self, gpx_files, photo_list, file_path, order, image_size,
                   title, track_color, track_width, between_color,
                   between_width, clickable, show_waypoints, show_track,
                   show_between_tracks):

    """ Write the KMZ file with the specified options. """

    kmz = kmlhandler.KMZ(file_path, self.logger, gpx_files,
                         photo_list, order, image_size, title, track_color,
                         track_width, between_color, between_width, clickable)

    kmz.writeFile(write_photos = True,
                  write_waypoints = show_waypoints, write_track = show_track,
                  write_between_tracks = show_between_tracks)

