# This file belongs to Happy Camel.
#
# Copyright (C) 2006-2008 Pieter Edelman (p _dot_ edelman AT gmx _dot_ net)
# with contributions by Tim Cera and Steffen Siebert.
#
# Happy Camel is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published
# by the Free Software Foundation; version 2 only.
#
# Happy Camel is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#

""" Module to handle photo files on the hard disk. There are two classes, Photo
    and PhotoList. The former one is a class for reading and writing each
    individual photo. The latter one is a collection of photo objects, with
    methods for certain batch operations. It inherits from a normal Python list.
"""

# There are different handlers for different image formats
# Metadata reading and writing by The Big Picture (fast)
TBP_READ_FORMATS =       ["jpg", "jpeg", "tif", "tiff"]
TBP_WRITE_FORMATS =      ["jpg", "jpeg", "tif", "tiff"]
# Metadata reading and writing by Exiftool (slower)
EXIFTOOL_READ_FORMATS  = ["acr", "ai", "arw", "bmp", "dib", "cr2", "cr2",
  "ciff", "dcm", "dc3", "dic", "dicm", "dng", "erf", "fpx", "gif", "jp2", "jpx",
  "jpeg", "jpg", "miff", "mif", "mos", "mrw", "nef", "orf", "pdf", "pef",
  "pict", "pct", "png", "jng", "mng", "ppm", "pbm", "pgm", "ps", "eps", "epsf",
  "psd", "qtif", "qti", "qif", "raf", "raw", "sr2", "srf", "thm", "tiff", "tif",
  "vrd", "wdp", "x3f", "xmp"]
EXIFTOOL_WRITE_FORMATS = ["jpg", "jpeg", "tiff", "tif", "cr2", "crw", "ciff",
  "dng", "erf", "gif", "mos", "mrw", "nef", "orf", "pef", "png", "jng", "mng",
  "ppm", "pbm", "pgm", "ps", "eps", "epsf", "psd", "thm", "vrd", "wdp", "xmp"]
# Image reading and resizing
PIL_READ_FORMATS =       ["bmp", "cur", "dcx", "fpx", "gbr", "gd", "gif", "im",
  "imt", "jpeg", "jpg", "mic", "msp", "pcd", "pcx", "png", "ppm", "psd", "sgi",
  "tga", "tiff", "tif", "wal", "xbm", "xpm"]

import os, time, re, types, subprocess, StringIO, urllib2, threading, Queue, sys, glob
import geonames, TBP, gpxhandler, flickrlib

# Try to import the Pyton Imaging Library
try:
  import Image
except ImportError:
  CAN_RESIZE_PHOTOS = False
else:
  CAN_RESIZE_PHOTOS = True

def haveExiftool():
  """ Checks to see if Exiftool is available on the system. Returns True or
      False. """
  
  try:
    # First, try a cached response. That saves the expensive calling of an
    # external command.
    global EXIFTOOL_AVAILABLE
    exiftool_available = EXIFTOOL_AVAILABLE
  except NameError:
    # If we don't have a cached response, go figure it out
    
    exiftool_available = False 
    try:
      # Try opening a subprocess to exiftool, silencing it by redirecting its
      # output to a silent pipe.
      f = subprocess.Popen(["exiftool", "-ver"], stdout = subprocess.PIPE)
      status = f.wait()
      if (status == 0):
        exiftool_available = True
    except OSError:
      # When an OSError occurs, the program is not available
      exiftool_available = False
    
    # Cache the response
    EXIFTOOL_AVAILABLE = exiftool_available
    
  return exiftool_available
  
# The Flickr API key for Happy Camel
# WARNING: This key is assigned by Flickr to Pieter Edelman ("Mr. Pi") for use
# in Happy Camel. It may not be used by anyone else for anything else than
# Happy Camel. If you would like to base your code on Happy Camel, please apply
# for your own Flickr API key (http://www.flickr.com/services/api/keys/)
API_KEY    = "f6440367c161bad9248168bd0bbb0010"
API_SECRET = "0e5b916f8545d4db"

class Photo:
  """ Base class for reading and writing photo metadata. """

  def __init__(self, logger, time_delta = 0):
    """ Initialize a logger object. The optional time_delta specifies the number
        of seconds the camera clock runs before real time. """

    self.logger     = logger
    self.setTimeDelta(time_delta)

    # Various metadata fields
    self.date       = None
    self.comment    = None
    self.lat        = None
    self.lon        = None
    self.ele        = None
    self.country    = None
    self.city       = None
    self.sub_city   = None

  def __repr__(self):
    return self.getName()
  
  def setTimeDelta(self, time_delta):
    """ Set the number of seconds the camera clock runs before real time. """

    self.time_delta = time_delta

  def getDate(self):
    """ Return the date of the photo as number of seconds since the epoch, or
        None if no date was found. """

    if (self.date):
        return self.date - self.time_delta
    else:
        return None

  def getLocation(self):
    """ Return a list of latitude, longitude and elevation of the photo. If
        elevation is unknown, it is set tot None. If latitude and/or longitude
        are unknown, this method returns None. """

    if ((self.lat) and (self.lon)):
      return [self.lat, self.lon, self.ele]

  def getLat(self):
    """ Return the latitude of the photo as float, or None if it doesn't exist.
    """
    return self.lat

  def getLon(self):
    """ Return the longitude of the photo as float, or None if it doesn't exist.
    """
    return self.lon

  def getComment(self):
    """ Return the comment associated with the picture, or None if no comment
        was found. """
    return self.comment

  def setLocation(self, lat, lon, ele = None, overwrite = True):
    """ Set the GPS coordinates of the photo. lat and lon are floats of
        latitude and longitude. ele is the elevation in meters. If it is
        omitted, no elevation value is written to the photo. If overwrite is
        True, the GPS coordinates currently present in the metadata are
        overwritten, otherwise they are left intact. """

    if ((not self.lat) or overwrite):
      self.lat = lat
    if ((not self.lon) or overwrite):
      self.lon = lon
    if ((not self.ele) or overwrite):
      self.ele = ele

  def __scalePhoto__(self, file_obj, max_side):
    """ Scale the photo referred to by the given file_obj so that the longest
        size is max_side. If the original is smaller than the required format,
        it isn't scaled. Return a JPEG-encoded string with data. """

    try:
      # Try to open the file
      photo_data = Image.open(file_obj)
    except Exception, message:
      self.logger.message("The PIL couldn't read photo %s so I won't create a thumbnail of it" % self.getName())
      self.logger.debug("Here's the error: %s" % message)
      return False
    else:
      # Calculate the scaling factor
      photo_size = photo_data.size
      if (photo_size[0] > photo_size[1]):
        fraction = float(max_side) / photo_size[0]
      else:
        fraction = float(max_side) / photo_size[1]
      
      # Only scale when the original is larger than the required size
      if (fraction < 1.0):
        # Scale the photo using the PIL thumbnail() function. This is much more
        # efficient then direct scaling functions.
        thumbnail_size = [int(photo_size[0] * fraction), int(photo_size[1] * fraction)]
        self.logger.debug("Resizing photo %s from %dx%d to %dx%d pixels." % (self.getName(), photo_size[0], photo_size[1], thumbnail_size[0], thumbnail_size[1]))
        photo_data.thumbnail(thumbnail_size, Image.BICUBIC)
      else:
        self.logger.debug("Required thumbnail for %s is larger than the original - no scaling." % self.getName())
        
      # Return a JPEG encoded string
      return photo_data.tostring("jpeg", "RGB")

    return False

  def __parseDate__(self, date_str):
    """ Convert a datetime string formatted according to YYYY.MM.DD HH.MM.SS
        into the number of seconds since the epoch, where "." means any
        character. """
    year   = int(date_str[0:4])
    month  = int(date_str[5:7])
    day    = int(date_str[8:10])
    hour   = int(date_str[11:13])
    minute = int(date_str[14:16])
    second = int(date_str[17:19])
    return time.mktime([year, month, day, hour, minute, second, 0, 0, time.timezone])

  def __DMS2D__(self, fracs):
    """ Convert a list of degrees, minutes, seconds into one degrees float. """
    return fracs[0] + ((1/60.0) * fracs[1]) + ((1/3600.0) * fracs[2])

class PhotoList(list):
  """ Control and manipulate a collection of Photo objects, which may be either
      photos on disk or photos on Flickr. This class also acts as a standard
      Python list. """

  # The protocols that we support
  PROTOCOLS = ["file", "flickr"]

  # The number of threads we want for disk operations and Flcikr calls
  DISK_THREAD_NUM   = 3
  FLICKR_THREAD_NUM = 10 # We can hit Flickr pretty hard, it's built for it

  def __init__(self, photos, time_delta, logger,
               read_tbp = True, write_tbp = False, flickr_api = None):
    """ Initialize the list with a list of photo specifications, which may be
        file names of Flickr photos. File names may be both absolute and
        relative. Flickr photos may be specified individually with the protocol
        flickr://photo-id or using a set with flickr://sets/set-id.
        time_delta specifies the number of seconds the camera clock runs before
        real time. read_tbp and write_tbp specify whether to use the TBP library
        instead of Exiftool for respectively reading and writing of metadata for
        photos on disk.
        flickr_api is the FlickrAPI object for communicating with Flickr. """

    self.logger    = logger
    self.read_tbp  = read_tbp
    self.write_tbp = write_tbp

    list.__init__(self)
    self.time_delta = time_delta

    curr_dir = os.getcwd()
    for photo in photos:
      # Figure out the protocol
      parts = photo.split("://")
      if (len(parts) == 2):
        protocol, path = parts
      else:
        protocol = "file"
        path     = parts[0]

      # A protocol may be specified with only the first part, but then it
      # shouldn't be ambiguous. Test whether the protocol is supported and
      # unambiguous.
      supported = False
      for test_protocol in self.PROTOCOLS:
        if (protocol == test_protocol[:len(protocol)]):
          if (supported):
            self.logger.error("Protocol \"%s\" is ambiguous (could both be \"%s\" and \"%s\")!" % (protocol, full_protocol, test_protocol))
            sys.exit(1)
          supported     = True
          full_protocol = test_protocol
      protocol = full_protocol
      if (not supported):
        self.logger.error("Protocol \"%s\" is not supported" % protocol)

      # Load photos from disk
      if (protocol == "file"):
        # Do the Unix-style globbing if it's not done so already by the shell
        expanded_path = glob.glob(os.path.expanduser(path))
        if (len(expanded_path) == 0):
          self.logger.error("\"%s\" doesn't exist." % path)
        for expanded_path in expanded_path:
          file_path = os.path.abspath(os.path.join(curr_dir, expanded_path))

          # imghdr doesn't seem to recognize all jpegs, so we use the extension
          extension = os.path.splitext(file_path)[1][1:].lower() # The extension includes the dot, so we have to chop this off
          if (read_tbp) and (extension in TBP_READ_FORMATS):
            self.append(DiskPhoto(file_path, self.logger, self.time_delta,
                                  read_tbp = True, write_tbp = write_tbp))
          elif (extension in EXIFTOOL_READ_FORMATS) and (haveExiftool()):
            self.append(DiskPhoto(file_path, self.logger, self.time_delta,
                                  read_tbp = False, write_tbp = False))
      # Load photos from Flickr
      elif (protocol == "flickr"):
        
        # Keep track of whether the protocol is well-formed
        well_formed = False
        if (path[:5] == "sets/"):
          # We're dealing with a photo set
          match = re.search("sets/(.+?)/?$", path)
          if (match):
            well_formed = True
            path = match.group(1)
            
            try:
              # If we can convert it directly to an integer, the path is a
              # set-id
              path = int(path)
            except ValueError:
              # Otherwise, it's a set name we have to search for.
              
              # At the moment, we haven't found anything yet, so we set the path
              # to None. For searching case insensitive, use a lowercase version
              # of the path.
              set_name = path.lower()
              path     = None

              # We can only search for sets in our own account.
              if not (flickr_api.getUser()):
                self.logger.error("Searching Flickr sets by name is only supported for logged in users - discarding %s." % photo)
              else:
                # Get all the sets for the user and loop over them to find the
                # appropiately named set.
                result = flickr_api.call("photosets.GetList", {})
                set_id = []
                for set in result["photosets"]["photoset"]:
                  if (set["title"]["_content"].lower() == set_name):
                    set_id.append(int(set["id"]))
                    
                # If we have exactly one result, we found it.
                if (len(set_id) > 1):
                  self.logger.error("Set \"%s\" occurs multiple times on Flickr-account %s - ignoring." % (set_name, flickr_api.getUser()))
                elif (len(set_id) == 1):
                  path = set_id[0]
                  self.logger.debug("Found set id %d for set \"%s\"" % (path, set_name))
                else:
                  well_formed = False
            
            if (path):
              # In rare cases, photosets contain more than 500 images.
              # Flickr returns 500 photo results per page, so we have to loop
              # over all pages to load all photos.
              page = total_pages = 1
              while (page <= total_pages):
                try:
                  result = flickr_api.call("photosets.getPhotos",
                                          {"photoset_id": path, "page": page})
                  page += 1
                  total_pages = int(result["photoset"]["pages"])
                  # Append each individual photo in the set to the list for, so it
                  # will later be processed.
                  for id in result["photoset"]["photo"]:
                    photos.append("flickr://" + id["id"])
                except flickrlib.FlickrFailure, inst:
                  # If Flickr can't find the photoset, it returns error 1.
                  if (inst.num == 1):
                    well_formed = False
                    break
                  else:
                    raise
        else:
          # Handle individual photos
          match = re.search("^([0-9]+)", path)
          if (match):
            well_formed = True
            path = int(match.group(1))
            self.append(FlickrPhoto(path, flickr_api, self.logger,
                                    self.time_delta))
        # Let the user now when she gave a wrong specification.
        if (not well_formed):
          logger.error("Can't make sense of \"%s\"" % photo)

  def setTimeDelta(self, time_delta):
    """ Set the number of seconds the camera clock runs before real time for
        every photo in the list. """

    self.time_delta = time_delta
    for photo in self:
      photo.setTimeDelta(self.time_delta)
  
  def getTimeDelta(self):
    """ Return the general number of seconds the camera clock runs before real
        time for every photo in the list. """
    
    return self.time_delta

  def __runThreadInstance__(self, method, pool, results, kwargs):
    """ Helper method for concurrent execution. It is not meant to be called
        directly, only through a threaded.Thread instance.
        It processes the objects in the pool Queue.Queue instance, each time
        calling the specified method on it with the provided keyword arguments.
        The results are stored in the results queue, as lists containg the
        object and the result. """

    exceptions = []
    
    # Process untill the queue gets empty
    while True:

      # Try to read a work object from the pool
      try:
        work = pool.get(block = False)
      except Queue.Empty:
        work = None

      # If we got an object, process it
      if (work):
        # It is absolutely necesseray to fill the queue with as many results as
        # the objects in the queue, because queue processors might rely on it.
        # So if an error occurs in the processing, we capture it, put an empty
        # result in the queue, and raise everything after we're done processing.
        try:
          result = eval("work.%s(**kwargs)" % method)
          results.put([work, result])
        except Exception, inst:
          results.put([work, False])
          exceptions.append(inst)
      # If we didn't get an object, we're done
      else:
        break
    
    for exception in exceptions:
      raise exception

  def __runThreaded__(self, num_threads, method, pool, results = Queue.Queue(),
                      **kwargs):
    """ Run the method method of the objects in the list pool concurrently and
        return a Queue.Queue object with the results as [work, result].
        pool may be a list, tuple, or Queue.Queue instance.
        num_threads specifies the maximum number of threads to use.
        The results of the child method can be accessed by providing a reference
        to a results queue. If you choose to consume the queue yourself, this
        method will return an empty queue at the end of the day.
        Arbitrary keyword arguments may be added to the call. """

    # The concurrent execution is obtained by calling a single method
    # in multithreaded fashion. Each of the concurrent instances receives the
    # same Queue.Queue instance, so together they can consume it. A results 
    # queue is provided to store the results. This is done as list of the Photo
    # instance and the actual result. After processing, it is converted to a
    # dict and returned.
    # If the user provided a list or tuple, convert it to a queue.
    if type(pool) in [types.ListType,  types.TupleType]:
      queue_pool = Queue.Queue()
      for work in pool:
        queue_pool.put_nowait(work)
      pool = queue_pool

    # Start the num_threads number of threads and store them in a list
    threads = []
    for thread_num in range(num_threads):
      thread = threading.Thread(target = self.__runThreadInstance__,
                                args = [method, pool, results, kwargs])
      thread.start()
      threads.append(thread)

    # Now block until all threads have finished
    for thread_num in range(num_threads):
      threads[thread_num].join()

    return results

  def __queue2Dict__(self, queue):
    """ Convert a Queue.Queue fifo as returned by __runThreaded__() to a dict,
        where the keys are the objects and the values the returned results. 
        """
    ret_dict = {}
    while True:
      try:
        result = queue.get(block = False)
        ret_dict[result[0]] = result[1]
      except Queue.Empty:
        break
    return ret_dict

  def getByType(self):
    """ Return a dict with lists of the different kind of photo objects. The
        keys are "disk" and "flickr". """
    
    photos = {"disk": [], "flickr": []}
    for photo in self:
      if isinstance(photo, DiskPhoto):
        type = "disk"
      else:
        type = "flickr"
      photos[type].append(photo)

    return photos
    
  def loadMetadata(self, read_date, read_comment, read_pos, read_flickr_data,
                   flickr_sources, time_delta):
    """ Load the metadata of all photos. Return a dictionary with the photo
        objects as keys and a bool indicating success as value. """

    # The processing of Flickr photos means a lot of http calls with little data
    # exchange, so it screams for concurrency. Also disk photos may be processed
    # in a multithreaded fashion for speed improvements, but with a different
    # number of threads.
    pools = self.getByType()
    
    # Process the pools
    disk_results =   self.__runThreaded__(self.DISK_THREAD_NUM, "loadMetadata",
                                          pools["disk"], read_date = read_date,
                                          read_comment = read_comment,
                                          read_pos = read_pos,
                                          time_delta = time_delta)
    flickr_results = self.__runThreaded__(self.FLICKR_THREAD_NUM, "loadMetadata",
                                          pools["flickr"], read_date = read_date,
                                          read_comment = read_comment,
                                          read_pos = read_pos,
                                          read_flickr_data = read_flickr_data,
                                          sources = flickr_sources,
                                          time_delta = time_delta)

    # Store all the results in a single dict and return it
    results = self.__queue2Dict__(disk_results)
    results.update(self.__queue2Dict__(flickr_results))
    return results

  def writeMetadata(self, write_coords = True, write_geo_names = False,
                    correct_time = False):
    """ Write the metadata to all photos. Return a dictionary with the photo
        objects as keys and a bool indicating success as value. """

    # The processing of Flickr photos means a lot of http calls with little data
    # exchange, so it screams for concurrency. Also disk photos may be processed
    # in a multithreaded fashion for speed improvements, but with a different
    # number of threads.
    pools = self.getByType()
    
    # Process the pools
    disk_results =   self.__runThreaded__(self.DISK_THREAD_NUM, "writeMetadata",
                                          pools["disk"],
                                          write_coords = write_coords,
                                          write_geo_names = write_geo_names,
                                          correct_time = correct_time)
    flickr_results = self.__runThreaded__(self.FLICKR_THREAD_NUM, "writeMetadata",
                                          pools["flickr"],
                                          write_coords = write_coords,
                                          write_geo_names = write_geo_names,
                                          correct_time = correct_time)

    # Store all the results in a single dict and return it
    results = self.__queue2Dict__(disk_results)
    results.update(self.__queue2Dict__(flickr_results))
    return results
    
  def getThumbnailData(self, max_side, queue):
    """ Load all the thumbnails from the images in the list and put a tuple of
        (photo object, JPEG-encoded thumbnail data) in the provided Queue.queue
        instance. This allows for a multithreaded processing of the images.
        max_side is the number of pixels the largest side of the thumbnail
        should have. """

    # The processing of Flickr photos means a lot of http calls with little data
    # exchange, so it screams for concurrency. Also disk photos may be processed
    # in a multithreaded fashion for speed improvements, but with a different
    # number of threads.
    pools = self.getByType()
    
    # Process the pools
    self.__runThreaded__(self.DISK_THREAD_NUM, "getThumbnailData",
                         pools["disk"], queue, max_side = max_side)
    self.__runThreaded__(self.FLICKR_THREAD_NUM, "getThumbnailData",
                         pools["flickr"], queue, max_side = max_side)

  def cleanup(self, check_date = False, check_location = False):
    """ Removes the photos from the list which don't comply to the specified
        criterium. Return a list of photo's that has been removed. """

    removed_photos = PhotoList([], self.time_delta, self.logger)

    photo_num = 0
    while (photo_num < len(self)):
      if ((check_date) and not (self[photo_num].getDate())):
        self.logger.message("No date found for photo %s" % self[photo_num].getName())
        removed_photos.append(self[photo_num])
        del self[photo_num]
      elif ((check_location) and not (self[photo_num].getLocation())):
        self.logger.message("No position found for photo %s" % self[photo_num].getName())
        removed_photos.append(self[photo_num])
        del self[photo_num]
      else:
        photo_num += 1

    self.logger.debug("Photolist after cleanup:")
    self.logger.debug(str(self))

    return removed_photos

  def getBeginEnd(self):
    """ Return the first and last date of the photo's in the list. """
    
    self.sortByDate()
    return (self[0].getDate(), self[-1].getDate())

  def queryGeoNames(self, radius, info = None):
    """ Query the Geonames.org service to get the place names for all photo's in
        the list. radius specifies the maximum amount of kilometers the place
        may be away from the coordinates to still be considered the place name.
        info is a string of tokens that determines which info needs to be
        queried; "n" means nation name, "c" means city name, "s" means subcity
        name. """

    # Make a list of photos to include and a list of their coordinates
    photos      = []
    coordinates = []
    for photo in self:
      if ((photo.lat) and (photo.lon)):
        photos.append(photo)
        coordinates.append([photo.lat, photo.lon])

    # Check what kind of info we need to query
    if (info):
      info = info.lower()
      if ("n" in info): query_country   = True
      else:             query_country   = False
      if ("c" in info): query_place     = True
      else:             query_place     = False
      if ("s" in info): query_sub_place = True
      else:             query_sub_place = False

    # Call geonames
    gn = geonames.Geonames()
    places = gn.getPlaces(coordinates, radius, query_country, query_place, query_sub_place)

    # Set the geo info to the photos
    for index in range(len(photos)):
      if places[index]:
        place = places[index]
        photo = photos[index]
        # Although it is possible to use extended character sets in IPTC data,
        # this is not widely supported by all applications. Therefore, we
        # convert the unicode strings from geonames.org to plain ASCII. Special
        # characters are replaced by a question mark.
        if (query_country):
          if (place["country_name"]):
            photo.country = place["country_name"].encode("ascii", "replace")
        if (query_place) and (place["place"]):
          photo.city = place["place"].encode("ascii", "replace")
        if (query_sub_place) and (place["sub_place"]):
          photo.sub_city = place["sub_place"].encode("ascii", "replace")

  def matchTracks(self, loc_sources, elevation_mode = 1, max_deviation = 0,
                  match_outside = False):
    """ Match the photos in the list to the location sources. loc_sources is
        a list of location sources, which can be gpxhandler.GPXFile objects or
        True for indicating that position data already present should be
        preserved. The most important source should be at the front of the list.
        elevation_mode specifies how elevation should be handled (0 = no
        elevation, 1 = only exact matches, 2 = interpolate).
        max_deviation specifies the deviation in meters that a photo may be
        interpolated between tracks.
        match_outside is a bool that determines if the position of photos made
        before or after the tracklogs start should be set to the lower and
        upper values of the tracks.
    """

    if (len(loc_sources)):
      self.logger.statement("Looking up positions")

      # Make sure we're processing in the order from less to most important
      loc_sources.reverse()

      # Store the very first and very last trackpoint of all the tracks, in
      # order of importance
      first_date = None
      last_date  = None
      overwrite_before = False
      overwrite_after  = False

      # Iterate over all the input sources
      for loc_source_num in range(len(loc_sources)):
        loc_source = loc_sources[loc_source_num]

        # If we encounter a track, match it
        if (isinstance(loc_source, gpxhandler.GPXFile)):
          # If we encounter True further along the way, don't overwrite any
          # position values already present in the exif data.
          if (True in loc_sources[loc_source_num:]):
            overwrite = False
          else:
            overwrite = True

          # Do the matching
          loc_source.matchTrack(self, elevation_mode, max_deviation,
                                overwrite = overwrite)

          # Keep track of the first and last points of all tracks
          if (match_outside):
            track_extremes = loc_source.getBeginEnd()
            if ((not first_date) or (track_extremes[0]["date"] <= first_date["date"])):
              first_date = track_extremes[0]
              overwrite_before = overwrite
            if ((not last_date) or (track_extremes[1]["date"] >= last_date["date"])):
              last_date = track_extremes[1]
              overwrite_after = overwrite

      # If needed, match photos before and after the tracks
      # TODO: What about overwrite??
      if (match_outside):
        for photo in self:
          if (photo.getDate() < first_date["date"]):
            self.logger.debug("Setting photo '%s' to tracks start (overwrite = %s)" % (photo.getName(), overwrite_before))
            photo.setLocation(first_date["lat"], first_date["lon"], None, overwrite_before)
          elif (photo.getDate() > last_date["date"]):
            self.logger.debug("Setting photo '%s' to tracks end (overwrite = %s)" % (photo.getName(), overwrite_after))
            photo.setLocation(last_date["lat"], last_date["lon"], None, overwrite_after)

  def sortByDate(self, reverse = False):
    """ Sort the photo list by date. If reverse is True, the latest dates are
        put in front. """
    self.sort(key = Photo.getDate, reverse = reverse)

  def sortByLat(self, reverse = False):
    """ Sort the photo list by latitude. If reverse is True, the greatest
        latitudes are put in front. """
    self.sort(key = Photo.getLat, reverse = reverse)

  def sortByLon(self, reverse = False):
    """ Sort the photo list by longitude. If reverse is True, the greatest
        longitudes are put in front. """
    self.sort(key = Photo.getLon, reverse = reverse)

class DiskPhoto(Photo):
  """ Read and write metadata of a photo on disk. """

  def __init__(self, file_path, logger, time_delta = 0,
               read_tbp = True,  write_tbp = False):
    """ Initialize with a full path of the photo on disk and a logger object.
        The optional time_delta specifies the number of seconds the camera clock
        runs before real time. If read_tbp and write_tbp are True, the TBP
        library is used for respectively reading and writing instead of the
        external Exiftool command. """

    # Initialize the base class
    Photo.__init__(self, logger, time_delta)

    # The path on disk
    self.file_path = file_path

    # If we can load an image using TBP, we need to save it for later
    self.read_tbp  = read_tbp
    self.write_tbp = write_tbp
    self.tbp_image = None

  def getName(self):
    """ Return the file name (leaf of the path) on disk of the photo file. """

    return os.path.basename(self.file_path)

  def getPath(self):
    """ Return the full path on disk of the photo. """

    return self.file_path

  def getURL(self):
    """ Return the URL linking to the photo. """

    return "file://%s" % os.path.join(os.getcwd(), self.getPath())

  def getThumbnailData(self, max_side = 200):
    """ Convert the image to a thumbnail and return the data, JPEG-encoded,
        or False if we cannot convert this image. max_side specifies the number
        of pixels the long side of the thumbnail needs to have. """

    # Check if we have the ability to write this format
    ext = os.path.splitext(self.getName())[1][1:].lower()
    if (ext in PIL_READ_FORMATS):
      self.logger.debug("Creating thumbnail for photo %s" % self.getName())
      photo_data = open(self.getPath())
      return self.__scalePhoto__(photo_data, max_side)
    else:
      self.logger.error("Can't create thumbnail for photo %s" % self.getName())
      return False

  def loadMetadata(self, read_date = False, read_comment = False,
                   read_pos = False, time_delta = 0):
    """ Extract the necessary metadata, which can be a combination of
        creation date, comment, and position. Return True on succes or
        False on failure. """

    date    = None
    comment = None
    lat     = None
    lon     = None
    ele     = None

    # Keep track of whether we were succesful reading the file
    success = False

    # Get the file extensions
    ext = os.path.splitext(self.getName())[1][1:].lower()

    # Try to load the file using TBP
    if (self.read_tbp) and (ext in TBP_READ_FORMATS):
      if (ext in ["jpg", "jpeg"]):
        self.tbp_image = TBP.Jpeg(self.file_path)
      elif (ext in ["tif", "tiff"]):
        self.tbp_image = TBP.Tiff(self.file_path)

      # Process the different types of data which needs to processed

      # Check both DateTimeOriginal and DataTimeDigitized (AKA CreateDate), in
      # case one of them isn't present.
      if (read_date):
        date = self.tbp_image.getExifTag("DateTimeOriginal", 2)
        if (not date):
          date = self.tbp_image.getExifTag("DateTimeDigitized", 2)
        if (date):
          date = self.__parseDate__(date)

      # Try to get a comment from the file
      # First try a JPEG comment
      if (read_comment):
        if (isinstance(self.tbp_image, TBP.Jpeg)):
          comments = self.tbp_image.getComments()
          # Search for the first valid comment
          if (comments):
            for comment in comments:
              # Some JPEG comments have null characters embedded, which need to
              # be stripped out
              comment = comment.replace("\x00", "")
              if (comment != ""):
                break
      # If there isn't a JPEG comment, try other tags
      # TODO: Usercomment needs special decoding powers. This should maybe
      # be moved to TBP
##      if (not comment):
##        comment = self.tbp_image.getExifTag("UserComment", 2)
      if ((not comment) or (comment == "")):
        comment = self.tbp_image.getExifTag("ImageDescription", 1)

      # Read the position data
      if (read_pos):
        lat     = self.tbp_image.getExifTag("GPSLatitude", 3)
        lat_ref = self.tbp_image.getExifTag("GPSLatitudeRef", 3)
        lon     = self.tbp_image.getExifTag("GPSLongitude", 3)
        lon_ref = self.tbp_image.getExifTag("GPSLongitudeRef", 3)
        ele     = self.tbp_image.getExifTag("GPSAltitude", 3)
        ele_ref = self.tbp_image.getExifTag("GPSAltitudeRef", 3)

        if ((lat) and (lat_ref)):
          lat = self.__DMS2D__(lat)
          if (lat_ref) and (lat_ref[0] == "S"): # Southern hemisphere
            lat = -1 * lat
        else:
          lat = None

        if ((lon) and (lon_ref)):
          lon = self.__DMS2D__(lon)
          if (lon_ref) and (lon_ref[0] == "W"): # Western hemisphere
            lon = -1 * lon
        else:
          lon = None

        if (ele != False):
          if (ele_ref == 1): # Below sea level is indicated by a 1
            ele = -1 * ele

      success = True

    # If we can't load the file natively, try to open it with Exiftool
    elif (ext in EXIFTOOL_READ_FORMATS) and (haveExiftool()):

      # Construct the command line to Exiftool
      exiftool_cl = ["exiftool", "%s" % self.file_path, "-S"]

      # Read both DateTimeOriginal and CreateDate (AKA DataTimeDigitized), in
      # case one of them isn't present
      if (read_date):
        exiftool_cl.append("-DateTimeOriginal")
        exiftool_cl.append("-CreateDate")

      # Check JPEG Comments, ImageDescription and UserComment
      if (read_comment):
        exiftool_cl.append("-Comment")
        exiftool_cl.append("-ImageDescription")
        exiftool_cl.append("-UserComment")

      # The GPS data
      if (read_pos):
        exiftool_cl.append("-c")
        exiftool_cl.append("%.6f")
        exiftool_cl.append("-GPSPosition")
        exiftool_cl.append("-GPSAltitude")
        exiftool_cl.append("-GPSAltitudeRef")

      # Run Exiftool and get its output
      exiftool = subprocess.Popen(exiftool_cl, stdout = subprocess.PIPE,
                                  stderr = subprocess.PIPE)
      exiftool_status = exiftool.wait()
      exiftool_out = exiftool.stdout.read()
      exiftool_err = exiftool.stderr.read()
      if (exiftool_status != 0):
        self.logger.debug("Exiftool couldn't read the EXIF data from the file %s" % self.getName())
        self.logger.debug("Here's its output:")
        self.logger.debug(exiftool_out)
        self.logger.debug("And here's its error:")
        self.logger.debug(exiftool_err)
        return False
      if (exiftool_out == ""):
        self.logger.debug("Exiftool couldn't read the EXIF data from the file %s" % self.getName())
        self.logger.debug("Here's its error:")
        self.logger.debug(exiftool_err)
        return False

      # Parse date
      date_match = re.search("DateTimeOriginal: (.*?)$", exiftool_out, re.M)
      if (not date_match):
        date_match = re.search("CreateDate: (.*?)$", exiftool_out, re.M)
      if (date_match):
        date = self.__parseDate__(date_match.group(1))

      # Parse comment (Comment > ImageDescription > UserComment)
      comment_match = re.search("Comment: (.+)$", exiftool_out, re.M)
      if not (comment_match):
        comment_match = re.search("ImageDescription: (.+)$", exiftool_out, re.M)
      if not (comment_match):
        comment_match = re.search("UserComment: (.+)$", exiftool_out, re.M)
      if (comment_match):
        comment = comment_match.group(1)

      # Parse GPS data
      gps_match          = re.search("GPSPosition: ([\d\.]*)\s*([NS])\s*,\s*([\d\.]*)\s*([EW])$", exiftool_out, re.M)
      altitude_match     = re.search("GPSAltitude: ([\d\.]*).*$", exiftool_out, re.M)
      altitude_ref_match = re.search("GPSAltitudeRef: (.*)$", exiftool_out, re.M)
      if (gps_match):
        lat = float(gps_match.group(1))
        lon = float(gps_match.group(3))
        if (gps_match.groups(2)[1] == "S"):
          lat *= -1
        if (gps_match.groups(2)[3] == "W"):
          lon *= -1
      if (altitude_match and altitude_ref_match):
        ele = float(altitude_match.group(1))
        if (altitude_ref_match.group(1) == "B"):
          ele *= -1

      success = True

    # Set the values to self
    if (date):
      self.date = date
      self.logger.debug("Found date %s for photo %s" % (self.getDate(), self.getName()))
    if (comment):
      self.comment = comment
      self.logger.debug("Found comment \"%s\" for photo %s" % (self.comment, self.getName()))
    if (lat) and (lon):
      self.lat = lat
      self.lon = lon
      self.logger.debug("Found position (%.6f, %.6f) for photo %s" % (self.lat, self.lon, self.getName()))
    if (ele):
      self.ele = ele
      self.logger.debug("Found altitude %.6f for photo %s" % (self.ele, self.getName()))

    return success

  def writeMetadata(self, write_coords = True, write_geo_names = False,
                    correct_time = False):
    """ Write the relevant data to the metadata fields in the image. This can
        apply to the coordinates, the city/country, and the clock offset from
        real time. The first parameter is a boolean which specifies whether to
        write the coordinates, the second should be False is city/country info
        shouldn't be written, or a combination of "l" (IPTC Location) and "k"
        (IPTC Keywords) for the target to write this info to. The third is a
        boolean whether to shift the times. """

    # Prepare the data
    if ((self.lat) and (self.lon) and write_coords):
      if (self.lat > 0):
        lat_ref = "N"
      else:
        lat_ref = "S"
      if (self.lon > 0):
        lon_ref = "E"
      else:
        lon_ref = "W"

    # Get the file extensions
    ext = os.path.splitext(self.getName())[1][1:].lower()

    # Try to write using TBP
    if ((self.write_tbp) and (ext in TBP_WRITE_FORMATS)):
      if ((self.lat) and (self.lon) and write_coords):
        # Latitude and longitude are stored as triplets of degrees, minutes,
        # seconds in rational format. However, we are free to set the degree
        # to high precision and leave the minutes and seconds zero.
        self.tbp_image.setExifTag("GPSVersionID", [2, 2, 0, 0], 3) # FIXME: TBP should take care of this
        self.tbp_image.setExifTag("GPSMapDatum", "WGS-84", 3)
        self.tbp_image.setExifTag("GPSLatitude", [abs(self.lat), 0, 0], 3)
        self.tbp_image.setExifTag("GPSLatitudeRef", lat_ref, 3)
        self.tbp_image.setExifTag("GPSLongitude", [abs(self.lon), 0, 0], 3)
        self.tbp_image.setExifTag("GPSLongitudeRef", lon_ref, 3)
        if (self.ele != None): # Explicit test because elevation can be 0
          if (self.ele < 0):
            ele_ref = 1
          else:
            ele_ref = 0

          self.tbp_image.setExifTag("GPSAltitude", abs(self.ele), 3)
          self.tbp_image.setExifTag("GPSAltitudeRef", ele_ref, 3)

      # Process the geonames info
      if ((self.country or self.city or self.sub_city) and write_geo_names):
        write_geo_names = write_geo_names.lower()
        self.tbp_image.setIPTCTag("ApplicationRecordVersion", 4) # FIXME: TBP should take care of this
        if ("l" in write_geo_names):
          if (self.country):
            self.tbp_image.setIPTCTag("Country-PrimaryLocationName", self.country)
          if (self.city):
            self.tbp_image.setIPTCTag("City", self.city)
          if (self.sub_city):
            self.tbp_image.setIPTCTag("Sub-location", self.sub_city)
        if ("k" in write_geo_names):
          iptc_keywords = self.tbp_image.getIPTCTag("Keywords")
          if (type(iptc_keywords) != types.ListType):
            iptc_keywords = [iptc_keywords]
          for keyword in [self.country, self.city, self.sub_city]:
            if (keyword) and (keyword not in iptc_keywords):
              self.tbp_image.appendIPTCTag("Keywords", keyword)

      # Time correction
      corr_date = time.strftime("%Y:%m:%d %H:%M:%S", time.localtime(self.getDate()))
      self.tbp_image.setExifTag("DateTimeOriginal", corr_date, 2)
      # The DateTimeDigitized specifies the date and time of original data
      # digitization and is also know as CreateDate.
      # FIXME: Account for IPTC tags
      create_date = self.tbp_image.getExifTag("DateTimeDigitized", 2)
      if (create_date):
        self.tbp_image.setExifTag("DateTimeDigitized", corr_date, 2)

      # Write the image
      self.tbp_image.rename(self.getPath() + "_original")
      self.tbp_image.writeFile(self.getPath())
      return True

    elif (ext in EXIFTOOL_WRITE_FORMATS):
      exiftool_cl = ["exiftool", "-v0", "-overwrite_original", "-GPSMapDatum=WGS-84"]

      # Construct the command line for the coordinates
      if ((self.lat) and (self.lon) and write_coords):
        # Construct the command line options for latitude and longitude
        exiftool_cl.append("-gps:GPSLatitude=%f" % abs(self.lat))
        exiftool_cl.append("-gps:GPSLatitudeRef=%s" % lat_ref)
        exiftool_cl.append("-gps:GPSLongitude=%f" % abs(self.lon))
        exiftool_cl.append("-gps:GPSLongitudeRef=%s" % lon_ref)

        # If elevation is known, add it the the command line for exiftool
        if (self.ele != None): # Explicit test because elevation can be 0
          if (self.ele < 0):
            ele_ref = "B"
          else:
            ele_ref = "A"
          exiftool_cl.append("-gps:GPSAltitudeRef=%s" % ele_ref)
          exiftool_cl.append("-GPSAltitude=%f" % abs(self.ele))

      # Construct the command line for the geonames info
      if ((self.country or self.city or self.sub_city) and write_geo_names):
        write_geo_names = write_geo_names.lower()
        if ("l" in write_geo_names):
          if (self.country):
            exiftool_cl.append("-IPTC:Country-PrimaryLocationName=%s" % self.country)
          if (self.city):
            exiftool_cl.append("-IPTC:City=%s" % self.city)
          if (self.sub_city):
            exiftool_cl.append("-IPTC:Sub-location=%s" % self.sub_city)
        if ("k" in write_geo_names):
          if (self.country):
            exiftool_cl.append("-Keywords+=%s" % self.country)
          if (self.city):
            exiftool_cl.append("-Keywords+=%s" % self.city)
          if (self.sub_city):
            exiftool_cl.append("-Keywords+=%s" % self.sub_city)

      # Construct the command line for time correction
      if (correct_time):
        if (self.time_delta < 0):
          exiftool_cl.append("-AllDates+=00:00:%d" % (-1 * self.time_delta))
        else:
          exiftool_cl.append("-AllDates-=00:00:%d" % self.time_delta)

      # Run exiftool
      if (len(exiftool_cl) > 4):
        # Run exiftool
        self.logger.debug("Writing location to the EXIF data of %s with arguments %s" % (self.getName(), exiftool_cl))
        exiftool = subprocess.Popen(exiftool_cl + [self.getPath()], stdout = subprocess.PIPE, stderr = subprocess.PIPE)
        exiftool_status = exiftool.wait()
        exiftool_out = exiftool.stdout.read()
        exiftool_err = exiftool.stderr.read()
        if ((exiftool_status != 0) or (exiftool_err != "")):
          self.logger.error("Exiftool returned an error when it tried to write the EXIF data to the file %s" % self.getPath())
          self.logger.error("Here's its output:")
          self.logger.error(exiftool_out)
          self.logger.error("And here's its error:")
          self.logger.error(exiftool_err)
          return False
        return True
    return False

class FlickrPhoto(Photo):
  """ Read and write metadata of a photo on Flickr. """

  def __init__(self, photo_id, flickr_api, logger, time_delta = 0):
    """ Initialize the photo with the photo_id on Flickr, a FlickrApi object,
        a logger object and a time_delta. """

    # Initialize the base class
    Photo.__init__(self, logger, time_delta)

    self.id     = photo_id
    self.flickr = flickr_api
    
    # Some other variables about the photos
    self.url    = None # The URL of the photo page
    self.title  = None # The photo title

  def getName(self):
    """ Return the photo id as string. """ # FIXME: append suffix

    if (self.title):
      return self.title
    return "%d" % self.id

  def getURL(self):
    """ Return the URL linking to the photo. """
    
    return self.url

  def loadMetadata(self, read_date = False, read_comment = False,
                   read_pos = False, read_flickr_data = False, 
                   sources = "fe", time_delta = 0):
    """ Download the metadata from Flickr, thus populating the object. The
        booleands read_date, read_comment and read_pos specify whether the date,
        comment and position should be read, respectively. Setting
        read_flickr_data to True forces the reading of extra Flickr data, like
        the photopage url. This is automatically done when reading other data
        from Flickr.
        sources is a string specifying one or more metadata sources to be
        processed: "f" meand Flickr, "e" means image Exif data. The source
        specified first takes the highest precedence. """

    # We store all the required information in these temporary bins.
    # We don't address the class variable untill the end of the
    # method, because these can allready hold a value from a previouse time,
    # but we need empty bins during the method to check for a succesfull read.
    date    = None
    comment = None
    lat     = None
    lon     = None
    ele     = None

    # At the end we need to say whether we succeeded.
    success = False

    # Check if we have enough metadata sources
    if (len(sources) == 0):
      raise ValueError, "No metadata sources specified. Choose at leas f or e."
    if (read_comment and ("f" not in sources)):
      raise "When reading comments you need Flickr as a metadata source, Exif alone is not sufficient."

    call_services = sources = sources.lower()
    if (read_flickr_data) and ("f" not in call_services):
      call_services += "f"
      
    # Check all different source types (Flickr data and/or Exif data)
    for service in call_services:
      if (service == "f"):
        # Load from Flickr

        # Call the photos.getInfo function for the current photo
        call = self.flickr.call("photos.getInfo", {"photo_id": self.id})
        if ("photo" in call):
          if ("f" in sources):
            # Extract date
            if ((read_date) and (not date)):
              try:
                date = self.__parseDate__(call["photo"]["dates"]["taken"])
              except KeyError:
                date = None

            # Extract description
            if (read_comment):
              try:
                comment = call["photo"]["description"]["_content"]
              except KeyError:
                comment = None

            # Extract loaction
            if ((read_pos) and (not (lat and lon))):
              try:
                lat = float(call["photo"]["location"]["latitude"])
                lon = float(call["photo"]["location"]["longitude"])
              except KeyError:
                lat = None
                lon = None

          # Extract some other useful information
          try:
            for url in call["photo"]["urls"]["url"]:
              if (url["type"] == "photopage"):
                self.url = url["_content"]
          except (KeyError, IndexError):
            self.url = None
            
          try:
            self.title = call["photo"]["title"]["_content"]
          except KeyError:
            self.title = None
            
      elif ((service == "e") and (read_pos or read_date)):
        # Load from Exif

        # Call the getExif method for the current photo
        call = self.flickr.call("flickr.photos.getExif", {"photo_id": self.id})
        date_original  = None
        date_digitized = None

        # We can't address the required tags directly, so we have to loop over
        # them to find the right one.
        for exif in call["photo"]["exif"]:
          label = exif["label"]
          value = exif["raw"]["_content"]

          if (label== "Date and Time (Original)"):
            date_original = value
          if (label == "Date and Time (Digitized)"):
            date_digitized = value
          if (label == "North or South Latitude"):
            if (value == "N"): lat_ref = 1
            else: lat_ref = -1
          if (label == "East or West Longitude"):
            if (value == "E"): lon_ref = 1
            else: lon_ref = -1
          if (label == "Longitude"):
            lon = value
          if (label == "Latitude"):
            lat = value

        # Set the values
        if ((read_date) and (not self.date)):
          if (date_original):
            date = self.__parseDate__(date_original)
          elif (date_digitized):
            date = self.__parseDate__(date_digitized)

        if ((read_pos) and (not (self.lat and self.lon))):
          if (lat and lat_ref and lon and lon_ref):
            lat = lat_ref * self.__parseCoordFracs__(lat)
            lon = lon_ref * self.__parseCoordFracs__(lon)

      # If we have all information we need, we can stop and save on a call to
      # Flickr
      if not ((read_date and not date) or (read_comment and not comment) or (read_pos and not (lat and lon))):
        success = True
        break

    # Set the variables we just found
    if (date):
      self.date = date
      self.logger.debug("Found date %s for photo %s" % (self.getDate(), self.getName()))
    if (comment):
      self.comment = comment
      self.logger.debug("Found comment \"%s\" for photo %s" % (self.comment, self.getName()))
    if (lat) and (lon):
      self.lat = lat
      self.lon = lon
      self.logger.debug("Found position (%.6f, %.6f) for photo %s" % (self.lat, self.lon, self.getName()))
    if (ele):
      self.ele = ele
      self.logger.debug("Found altitude %.6f for photo %s" % (self.ele, self.getName()))

    return True

  def getThumbnailData(self, max_side = 200):
    """ Convert the image to a thumbnail and return the data, JPEG-encoded,
        or False if we cannot convert this image. max_side specifies the number
        of pixels the long side of the thumbnail needs to have. """

    self.logger.debug("Creating thumbnail for photo %s" % self.getName())
    
    # Let's try to be smart and download an image from Flickr that already
    # matches the required image as close as possible (while still being
    # larger).
    # Ask Flickr what sizes it has.
    call = self.flickr.call("photos.getSizes", {"photo_id": self.id})
    
    # Loop over the available sizes and select the smallest one that still
    # is larger than the required size. For convenience, we store this
    # alongside the found size in the best_fit dict.
    # We alse search for the largest one a a backup option.
    best_fit = {"size": 10000000, "fake": True}
    largest  = {"size": 0}
    for entry in call["sizes"]["size"]:
      if (entry["label"] != "Square"): # Discard the square image
        width  = int(entry["width"])
        height = int(entry["height"])

        # Choose which side to consider (width or height)
        if (width > height):
          longest = width
        else:
          longest = height

        # If this one fits better, select it
        if ((longest < best_fit["size"]) and (longest >= max_side)):
          best_fit = {"size": longest, "entry": entry}
        if (longest > largest["size"]):
          largest = {"size": longest, "entry": entry}

    # If we couldn't find a matching picture, take the largest one available.
    if ("fake" in best_fit):
      best_fit = largest
      
    # Download the photo and load it as a PIL image
    photo_data = StringIO.StringIO(urllib2.urlopen(best_fit["entry"]["source"]).read())
    return self.__scalePhoto__(photo_data, max_side)

  def writeMetadata(self, write_coords = True, write_geo_names = False,
                    correct_time = False):
    """ Write the relevant data to the Metadata on Flickr. This can apply to the
        coordinates, the city/country, and the corrected datetimes. This is
        specified by the booleand write_coords, write_geo_names and correct_time
        respectively.
        Note: this method does not alter the data embedded in the photo, just
        the metadata on Flickr. """

    # FIXME: Check for success!
    # Write the coordinates using photos.geo.setLocation
    if (write_coords):
      call = self.flickr.call("photos.geo.setLocation",
                              {"photo_id": self.id,
                               "lat": self.lat, "lon": self.lon})

    # Write the geonames as tags
    if (write_geo_names):
      tags = ""
      for place in [self.country, self.city, self.sub_city]:
        if (place):
          if (len(tags)):
            tags += ", "
          tags += place

      call = self.flickr.call("photos.addTags",
                              {"photo_id": self.id, "tags": tags})

    # Write the corrected time using photo.setDates
    corr_date = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.getDate()))
    call = self.flickr.call("photos.setDates",
                            {"photo_id": self.id, "date_taken": corr_date})

    return True

  def __parseCoordFracs__(self, fracs_str):
    """ Convert a string of degrees, minutes, seconds returned by Flickr into
        one degrees float. """

    parse = re.search(r'([0-9\.]+)\\/1,\n([0-9\.]+)\\/1,\n([0-9\.]+)\\/1', fracs_str, re.S)
    if (parse):
      deg = float(parse.group(1))
      min = float(parse.group(2))
      sec = float(parse.group(3))
      return self.__DMS2D__([deg, min, sec])
    return False
